﻿#region Copyright
//
//  The token highlighting functionality for this class was based
//  on the following codeproject article:
//-----------------------------------------------------------------------
//
//  http://www.codeproject.com/KB/recipes/TinyPG.aspx
//
// Copyright 2008 - 2010 Herre Kuijpers - <herre.kuijpers@gmail.com>
//
// This source file(s) may be redistributed, altered and customized
// by any means PROVIDING the authors name and all copyright
// notices remain intact.
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED. USE IT AT YOUR OWN RISK. THE AUTHOR ACCEPTS NO
// LIABILITY FOR ANY DATA DAMAGE/LOSS THAT THIS PRODUCT MAY CAUSE.
//
//  SEE ALSO:
//  Parsing Expression Grammar Support for C# 3.0 Part 1 - PEG Lib and Parser:
//  http://www.codeproject.com/KB/recipes/grammar_support_1.aspx
//
//-----------------------------------------------------------------------
#endregion Copyright

using System;

using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Threading;
using System.Diagnostics;

using Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;

namespace Infragistics.SyntaxParsing.Samples
{
    #region ContextSwitchEventArgs
    
    // delegate for firing context switch events
    public delegate void ContextSwitchEventHandler(object sender, ContextSwitchEventArgs e);

    [Serializable]
    [ComVisible(true)]
    public class ContextSwitchEventArgs : EventArgs
    {
        public readonly ParseNode PreviousContext;
        public readonly ParseNode NewContext;

        // Summary:
        //     Initializes a new instance of the System.EventArgs class.
        public ContextSwitchEventArgs(ParseNode prevContext, ParseNode nextContext)
        {
            PreviousContext = prevContext;
            NewContext = nextContext;
        }
    }

    #endregion ContextSwitchEventArgs

    #region ParseTreeTextHighlighter
    /// <summary>
    /// Highlights text that was parsed by a Grammar into a ParseTree.
    /// </summary>
    public class ParseTreeTextHighlighter : TextHighlighter
    {
        private ParseNode currentContext = null;
        public event ContextSwitchEventHandler SwitchContext;

        private ParseTree parseTree;
        private Grammar grammar = null;

        private Dictionary<TerminalSymbol, Color> terminalSymbolColorTable = null;
        private Dictionary<SymbolGroup, Color> symbolGroupColorTable = null;
        private Dictionary<Color, int> idTable = null;

        public event ParseCompleteHandler ParseComplete;

        public ParseTreeTextHighlighter( RichTextBox textBox, Grammar grammar, ParseTree parseTree ) : base( textBox )
        {
            this.grammar = grammar;
            this.parseTree = parseTree;
        }

        public void AssignColorTables(
            Dictionary<TerminalSymbol, Color> terminalSymbolColorTable,
            Dictionary<SymbolGroup, Color> symbolGroupColorTable )
        {
            this.terminalSymbolColorTable = terminalSymbolColorTable;
            this.symbolGroupColorTable = symbolGroupColorTable;

            this.InitializeIDTable( this.AllColors );
        }

        private IEnumerable<Color> AllColors
        {
            get
            {
                List<Color> colors = new List<Color>();

                if ( this.terminalSymbolColorTable != null )
                {
                    foreach ( Color color in this.terminalSymbolColorTable.Values )
                    {
                        colors.Add( color );
                    }
                }

                if ( this.symbolGroupColorTable != null )
                {
                    foreach ( Color color in this.symbolGroupColorTable.Values )
                    {
                        colors.Add( color );
                    }
                }

                return colors;
            }
        }

        private void InitializeIDTable( IEnumerable<Color> colors )
        {
            if ( this.idTable == null )
                this.idTable = new Dictionary<Color,int>();
            else
                this.idTable.Clear();

            int id = 1;
            foreach ( Color color in colors )
            {
                if ( this.idTable.ContainsKey(color) == false )
                {
                    this.idTable.Add( color, id );
                    id += 1;
                }
            }
        }

        #region OnTextBoxSelectionChanged
        internal override void OnTextBoxSelectionChanged( int selectionStart )
        {
            if ( this.stateLocked != IntPtr.Zero )
                return;

            if ( this.SwitchContext == null )
                return;

            ParseNode newContext = this.GetCurrentContext( selectionStart );
            
            if ( this.currentContext == null ) 
                this.currentContext = newContext;

            if ( newContext == null )
                return;

            if ( newContext.Token.CharacterPosition != this.currentContext.Token.CharacterPosition)
            {
                this.SwitchContext.Invoke(this, new ContextSwitchEventArgs(currentContext, newContext));
                this.currentContext = newContext;
            }
        }
        #endregion OnTextBoxSelectionChanged

        #region GetCurrentContext
        /// <summary>
        /// Returns the section in which the user is currently editing
        /// </summary>
        public ParseNode GetCurrentContext( int selectionStart )
        {
            ParseNode node = this.FindNode( this.parseTree, selectionStart );
            return node;
        }
        #endregion GetCurrentContext

        #region FindNode
        private ParseNode FindNode( ParseTree tree, int startPos )
        {
            foreach ( ParseNode node in tree.Nodes )
            {
                ParseNode found = this.FindNode( node, startPos );
                if ( found != null )
                    return found;
            }

            return null;
        }

        private ParseNode FindNode( ParseNode node, int startPos )
        {
            if ( node == null )
                return null;

            if ( node.Token.CharacterPosition <= startPos &&
                 (node.Token.CharacterPosition + node.Token.Text.Length) >= startPos)
            {
                foreach ( ParseNode n in node.Nodes )
                {
                    if (n.Token.CharacterPosition <= startPos &&
                        (n.Token.CharacterPosition + n.Token.Text.Length) >= startPos)
                        return FindNode(n, startPos);
                }

                return node;
            }
            else
                return null;

        }
        #endregion FindNode

        #region HighlightTokens
        private void HighlightTokens( IEnumerable<ParseTreeTokenBase> tokens, StringBuilder sb )
        {
            foreach ( ParseTreeTokenBase token in tokens )
            {
                int id = 0;
                string tokenText = token.Text;

                TerminalSymbol ts = token.TerminalSymbol;
                if ( ts != null )
                {
                    //  Change case
                    if ( ts.Value.Equals(tokenText, StringComparison.InvariantCultureIgnoreCase) )
                        tokenText = ts.Value;

                    //  Check the TerminalSymbol table first, then the
                    //  SymbolGroup table; if either one has a color defined
                    //  for this TerminalSymbol, use it
                    Color color = Color.Empty;
                    SymbolGroup group = ts.Group;
                    if ( this.terminalSymbolColorTable != null )
                        this.terminalSymbolColorTable.TryGetValue(ts, out color);

                    if ( color.IsEmpty &&
                         group != null && this.symbolGroupColorTable != null )
                        this.symbolGroupColorTable.TryGetValue(group, out color);

                    //  Get the id of the color from the id table
                    if ( color.IsEmpty == false )
                        this.idTable.TryGetValue( color, out id );
                }

                //  BF 9/13/11 WEIRD
                //if ( id > 0 )
                if ( id >= 0 )
                    sb.Append( string.Format( @"{{\cf{0} ", id) );

                tokenText = TextHighlighter.EscapeForRtf( tokenText );

                sb.Append( tokenText );

                if ( id > 0 )
                    sb.Append( "}" );
            }
        }
        #endregion HighlightTokens

        #region AddRtfHeader
        protected override void AddRtfHeader( StringBuilder sb )
        {
            string fontName = this.textBox != null ? this.textBox.Font.Name : "Lucida Console";
            IEnumerable<Color> allColors = this.idTable != null ? this.idTable.Keys : null;
            string colors = TextHighlighter.SerializeColors( allColors );
            string header = TextHighlighter.RtfHeaderTemplate.Replace( "{0}", fontName );
            header = header.Replace( "{1}", colors );
            sb.Insert(0, header);
        }
        #endregion AddRtfHeader

        #region HighlightTextCore
        protected override void HighlightTextCore()
        {
            if ( this.parseTree == null ||
                 this.parseTree.Nodes == null ||
                 this.parseTree.Nodes.Count == 0  )
                return;

            StringBuilder sb = new StringBuilder();
            this.HighlightTokens(this.parseTree.Tokens, sb);

            // NOTE: if you do not need unicode characters and you need a little bit more performance
            // you can comment out the following line:
            //sb = Unicode(sb);     // <--- without this, unicode characters will be garbled after highlighting

            this.AddRtfHeader(sb);
            this.AddRtfEnd(sb);

            this.VerifyEncoding( sb );

            this.textBox.Rtf = sb.ToString();

            if ( this.parseTree != null && this.ParseComplete != null )
                this.ParseComplete.Invoke( this, new ParseCompleteEventArgs(this.parseTree) );
        }
        #endregion HighlightTextCore

        #region Parse
        protected override object Parse( string text )
        {
            return this.grammar.Parse( text );
        }
        #endregion Parse

        #region AssignParseResult
        protected override void AssignParseResult( object value )
        {
            this.parseTree = value as ParseTree;
        }
        #endregion AssignParseResult

        #region OnMessagesResumed
        protected override void OnMessagesResumed()
        {
            if ( this.parseTree != null && this.parseTree.HasErrors )
            {
                IEnumerable<ErrorToken> errors = this.parseTree.Errors;
                foreach( ErrorToken error in errors )
                {
                    int len = error.Text.Length > 0 ? error.Text.Length : 1;
                    this.Underline( error.CharacterPosition, len, false );
                }
            }
        }
        #endregion OnMessagesResumed
    }
    #endregion ParseTreeTextHighlighter

    #region ParseCompleteEventArgs
    
    public delegate void ParseCompleteHandler( object sender, ParseCompleteEventArgs e );

    public class ParseCompleteEventArgs : EventArgs
    {
        private ParseTree parseTree = null;

        public ParseCompleteEventArgs( ParseTree parseTree )
        {
            this.parseTree = parseTree;
        }

        public ParseTree ParseTree { get { return this.parseTree; } }
    }

    #endregion ParseCompleteEventArgs

}
