﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.IO;

using Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;

namespace Infragistics.SyntaxParsing.Samples
{
    #region frmEBNFEditor class
    public partial class frmEBNFEditor : CodeEditorBase
    {
        #region Constants
        internal const string APP_NAME = "Infragistics EBNF script editor";
        private const string BACKUP_SUFFIX = "64fd64fd-d115-4201-935f-7de572b7bf11";
        internal const string FILE_DIALOG_FILTER = "EBNF script files (*.bnf, *.ebnf)|*.bnf;*.ebnf";
        #endregion Constants

        #region Member variables

        private GrammarImporter importer = null;

        private bool updatePending = true;

        private ImportResult lastResult = null;

        private EBNFTextHighlighter highlighter = null;

        private frmStatementCompletion statementCompletionForm = null;

        private List<string> statementCompletionExclusions = null;

        private Regex regexBackToNonWhitespace = new Regex( "[^\\s\n\t]", RegexOptions.RightToLeft );

        #endregion Member variables

        #region Constructor

        public frmEBNFEditor()
        {
            this.InitializeComponent();
            this.InitializeUI();
        }

        #endregion Constructor

        #region InitializeUI
        protected override void InitializeUI()
        {
            base.InitializeUI();

            this.Text = frmEBNFEditor.APP_NAME;
            this.KeyPreview = true;

            this.SetTextBoxText( string.Empty );

            this.saveToolStripMenuItem.Enabled = false;
            this.menuStrip1.Items.Add( this.toolsToolStripMenuItem );
            this.optionsToolStripMenuItem.Click += new EventHandler(this.OnMenuStripMenuItemClick);

            this.txtCode.MouseDown += new MouseEventHandler(this.txtCode_MouseDown);
            this.txtCode.MouseDoubleClick += new MouseEventHandler(txtCode_MouseDoubleClick);
            this.txtCode.KeyDown += new KeyEventHandler(this.txtCode_KeyDown);
            this.txtCode.KeyPress += new KeyPressEventHandler(this.txtCode_KeyPress);

            ContextMenu cm = new ContextMenu();
            MenuItem mi = null;
            mi = cm.MenuItems.Add( "Copy", new EventHandler( this.OnCodeEditorMenuItemClicked ) );
            mi.Name = mi.Text;
            mi = cm.MenuItems.Add( "Paste", new EventHandler( this.OnCodeEditorMenuItemClicked ) );
            mi.Name = mi.Text;
            mi = cm.MenuItems.Add("-");
            mi = cm.MenuItems.Add( "Go To Definition", new EventHandler( this.OnCodeEditorMenuItemClicked ) );
            mi.Name = "GoToDefinition";
            mi = cm.MenuItems.Add("-");
            mi = cm.MenuItems.Add( "Ignore Case", new EventHandler( this.OnCodeEditorMenuItemClicked ) );
            mi.Name = "IgnoreCase";
            mi = cm.MenuItems.Add( "Make Regular Expression", new EventHandler( this.OnCodeEditorMenuItemClicked ) );
            mi.Name = "MakeRegularExpression";
            mi = cm.MenuItems.Add("-");
            mi = cm.MenuItems.Add( "Rule Tree", new EventHandler( this.OnCodeEditorMenuItemClicked ) );
            mi.Name = "RuleTree";
            mi = cm.MenuItems.Add("-");
            mi = cm.MenuItems.Add( "Add Comment", new EventHandler( this.OnCodeEditorMenuItemClicked ) );
            mi.Name = "AddComment";

            this.txtCode.ContextMenu = cm;

            this.txtCode.Enabled = true;

            this.SetModified( false );

            this.RenameMenuItems( this.menuStrip1 );
        }

        #endregion InitializeUI

        #region Importer
        internal GrammarImporter Importer { get { return this.importer; } }
        #endregion Importer

        #region HasValidGrammar
        /// <summary>
        /// Returns a boolean value indicating whether the associated importer's
        /// Grammar is valid.
        /// </summary>
        public bool HasValidGrammar { get { return this.lastResult != null && this.lastResult.Success; } }
        #endregion HasValidGrammar

        #region Filename
        /// <summary>
        /// Returns the filename of the last grammar definition that was loaded
        /// </summary>
        public string Filename { get { return this.filename; } }
        #endregion Filename

        #region ShowMe
        public void ShowMe( frmMain mainForm, string filename )
        {
            Grammar grammar = null;
            this.ShowMe( mainForm, filename, grammar );
        }

        public void ShowMe( frmMain mainForm, string filename, Grammar grammar )
        {
            if ( this.importer == null )
                this.importer = new GrammarImporter();

            ImportResult result = null;

            this.highlighter = new EBNFTextHighlighter( this.txtCode, this.importer );
            this.highlighter.CompileComplete += new CompileCompleteHandler(this.highlighter_CompileComplete);
            this.highlighter.Highlighting += new EventHandler(this.highlighter_Highlighting);
            this.highlighter.HighlightComplete += new EventHandler(this.highlighter_HighlightComplete);
            
            if ( File.Exists(filename) )
            {
                Cursor.Current = Cursors.WaitCursor;

                this.filename = filename;
                string ebnf = File.ReadAllText( this.filename, frmMain.TextEncoding );
                this.SetTextBoxText( ebnf );
                this.txtCode.Enabled = true;
                this.closeToolStripMenuItem.Enabled = true;
                this.SetModified( false );

                Cursor.Current = Cursors.Default;
            }

            this.ShowDialog( mainForm );

            if( grammar != null && result != null && result.Success )
                grammar.InitializeFrom( importer.Grammar );
        }
        #endregion ShowMe

        #region OnActivated
        protected override void OnActivated(EventArgs e)
        {
            base.OnActivated(e);

            if ( updatePending )
                this.highlighter.HighlightText();
        }
        #endregion OnActivated

        #region highlighter_CompileComplete
        private void highlighter_CompileComplete(object sender, CompileCompleteEventArgs e)
        {
            IEnumerable<IErrorToken> errors = this.GetErrors(e.Result);
            this.LogResults( errors );
            this.toolTip.Enable( errors );

            this.SetTimeElapsed( -1 );

            this.toolTip.Enable( errors );
        }
        #endregion highlighter_CompileComplete

        #region highlighter_Highlighting
        void highlighter_Highlighting(object sender, EventArgs e)
        {
            this.suspended = true;
        }
        #endregion highlighter_Highlighting

        #region highlighter_HighlightComplete
        void highlighter_HighlightComplete(object sender, EventArgs e)
        {
            this.suspended = false;
        }
        #endregion highlighter_HighlightComplete

        #region OnMenuStripMenuItemClick
        private void OnMenuStripMenuItemClick(object sender, EventArgs e)
        {
            ToolStripMenuItem mi = sender as ToolStripMenuItem;
            string name = mi.Name;
            switch ( name )
            {
                case "options":
                    frmOptions options = new frmOptions();
                    options.ShowMe( this, this.importer );
                    break;
            }

            //base.OnMenuStripMenuItemClick( mi );
        }
        #endregion OnMenuStripMenuItemClick

        #region BackupScript
        private void BackupScript( string filename )
        {
            if ( string.IsNullOrEmpty(filename) ||
                 File.Exists(filename) == false )
                return;

            bool exists = true;
            int i = 0;
            string newFilename = null;
            while ( exists )
            {
                i += 1;
                newFilename = string.Format("{0}_{1}_{2}", filename, frmEBNFEditor.BACKUP_SUFFIX, i );
                exists = File.Exists(newFilename);
            }

            if ( string.IsNullOrEmpty(newFilename) == false )
            {
                File.Copy(filename, newFilename);
                FileAttributes attr = File.GetAttributes( newFilename );
                attr |= FileAttributes.ReadOnly;
                File.SetAttributes( newFilename, attr );
            }
        }
        #endregion BackupScript

        #region SetTextBoxText
        private void SetTextBoxText( string value )
        {
            this.txtCode.Text = value;
            this.SetModified( false );
        }
        #endregion SetTextBoxText

        #region OnCodeEditorMenuItemClicked
        private void OnCodeEditorMenuItemClicked(object sender, EventArgs e)
        {
            MenuItem mi = sender as MenuItem;
            NonTerminalSymbol nts = null;
            StatementToken token = null;

            switch ( mi.Name )
            {
                case "Copy":
                    this.OnMenuStripMenuItemClick( this.copyToolStripMenuItem, EventArgs.Empty );
                    break;

                case "Paste":
                    this.OnMenuStripMenuItemClick( this.pasteToolStripMenuItem, EventArgs.Empty );
                    break;

                case "GoToDefinition":
                    
                    SymbolToken symbolToken = mi.Tag as SymbolToken;
                    if ( symbolToken != null )
                    {
                        SymbolToken dec = this.importer.FindSymbolDeclaration( symbolToken.Symbol );
                        if ( dec != null )
                        {
                            this.txtCode.SelectionStart = dec.CharacterPosition;
                            this.txtCode.SelectionLength = dec.Length;
                        }
                    }
                    break;

                case "RuleTree":
                    
                    nts = mi.Tag as NonTerminalSymbol;
                    if ( nts != null )
                    {
                        frmRuleTree form = new frmRuleTree();
                        form.ShowMe( this, nts );
                    }

                    break;

                case "IgnoreCase":
                case "MakeRegularExpression":
                    TerminalSymbolToken tsToken = mi.Tag as TerminalSymbolToken;
                    int start = tsToken.CharacterPosition;
                    SpecialSequenceType sst = mi.Name == "IgnoreCase" ? SpecialSequenceType.TerminalSymbolLiteralIgnoreCase : SpecialSequenceType.TerminalSymbolRegex;
                    string newString = SpecialSequence.ToString( this.importer, sst );
                    int length = 0;

                    string s = this.txtCode.Text.Substring(0, tsToken.CharacterPosition);
                    Match m = this.regexBackToNonWhitespace.Match( s );
                    if ( m.Success )
                    {
                        SpecialSequence specialSequence = this.importer.FindToken(m.Index) as SpecialSequence;
                        if ( specialSequence != null )
                        {
                            start = specialSequence.CharacterPosition;
                            length = specialSequence.Text.Length;
                        }

                    }

                    this.txtCode.Select( start, length );
                    this.txtCode.SelectedText = newString;

                    break;

                case "AddComment":
                    int selStart = this.txtCode.SelectionStart;
                    token = this.importer.FindToken( selStart );
                    string pad = string.Empty;
                    if ( token != null && token.IsWhitespace == false )
                    {
                        selStart = token.CharacterPosition;
                        pad = " ";
                    }

                    string commentOpen = this.importer.Notation.CommentBlockOpeningSymbol;
                    string commentClose = this.importer.Notation.CommentBlockClosingSymbol;
                    string comment = "<comment>";
                    
                    this.txtCode.SelectionStart = selStart;
                    this.txtCode.SelectionLength = 0;
                    this.txtCode.SelectedText = string.Format( "{0} {1} {2}{3}", commentOpen, comment, commentClose, pad );
                    
                    selStart += commentOpen.Length + 1;
                    this.txtCode.SelectionStart = selStart;
                    this.txtCode.SelectionLength = comment.Length;

                    break;
            }
        }
        #endregion OnCodeEditorMenuItemClicked

        #region txtCode_MouseDown
        private void txtCode_MouseDown(object sender, MouseEventArgs e)
        {
            RichTextBox textBox = sender as RichTextBox;

            if ( e.Button == MouseButtons.Right )
            {
                MenuItem copy = textBox.ContextMenu.MenuItems["Copy"];
                MenuItem paste = textBox.ContextMenu.MenuItems["Paste"];
                MenuItem gotoDef = textBox.ContextMenu.MenuItems["GoToDefinition"];
                MenuItem ruleTree = textBox.ContextMenu.MenuItems["RuleTree"];
                MenuItem ignoreCase = textBox.ContextMenu.MenuItems["IgnoreCase"];
                MenuItem regex = textBox.ContextMenu.MenuItems["MakeRegularExpression"];
                MenuItem comment = textBox.ContextMenu.MenuItems["AddComment"];

                comment.Enabled = true;
                ignoreCase.Enabled = false;
                regex.Enabled = false;

                int selStart = textBox.SelectionStart;
                int selLength = textBox.SelectionLength;
                int charPos = textBox.GetCharIndexFromPosition( e.Location );
                bool hitSel = charPos >= selStart && (charPos <= (selStart + selLength) );
                if ( hitSel == false )
                {
                    textBox.SelectionStart = charPos;
                    textBox.SelectionLength = 0;
                }

                selStart = textBox.SelectionStart;
                selLength = textBox.SelectionLength;

                copy.Enabled = selLength > 0;
                paste.Enabled = string.IsNullOrEmpty( Clipboard.GetText() ) == false;

                SymbolToken token = null;
                token = this.importer.FindToken( textBox.SelectionStart ) as SymbolToken;
                NonTerminalSymbolToken ntsToken = null;

                bool enableGoToDefinition = false;
                if ( token != null )
                {
                    NonTerminalSymbolDeclaration ntsDec = token as NonTerminalSymbolDeclaration;
                    ntsToken = token as NonTerminalSymbolToken;
                    TerminalSymbolToken tsToken = token as TerminalSymbolToken;

                    if ( ntsDec != null )
                        enableGoToDefinition = false;
                    else
                    if ( ntsToken != null )
                        enableGoToDefinition = true;
                    else
                    if ( tsToken != null && tsToken.FromReference )
                        enableGoToDefinition = true;

                    if ( tsToken != null && tsToken.FromReference == false )
                    {
                        TerminalSymbol ts = tsToken.TerminalSymbol;
                        switch ( ts.Comparison )
                        {
                            case TerminalSymbolComparison.Literal:
                                ignoreCase.Enabled = true;
                                regex.Enabled = true;
                                ignoreCase.Tag = tsToken;
                                regex.Tag = tsToken;
                                break;

                            case TerminalSymbolComparison.LiteralIgnoreCase:
                                regex.Enabled = true;
                                regex.Tag = tsToken;
                                break;

                            case TerminalSymbolComparison.RegularExpression:
                                ignoreCase.Enabled = true;
                                ignoreCase.Tag = tsToken;
                                break;
                        }
                    }
                }

                gotoDef.Enabled = enableGoToDefinition;
                gotoDef.Tag = token;

                ntsToken = token as NonTerminalSymbolToken;
                ruleTree.Enabled = ntsToken != null;
                ruleTree.Tag = ntsToken != null ? ntsToken.NonTerminalSymbol : null;
            }
        }
        #endregion txtCode_MouseDown

        #region txtCode_MouseDoubleClick
        private void txtCode_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            RichTextBox textBox = sender as RichTextBox;
            int selStart = -1;
            int selLength = -1;

            StatementToken token = this.importer.FindToken( textBox.SelectionStart ); 
            if ( token != null )
            {
                selStart = token.CharacterPosition;
                selLength = token.Length;
            }
            else
            {
                string s = this.GetCurrentWord( textBox.SelectionStart, out selStart );
                if ( string.IsNullOrEmpty(s) == false )
                    selLength = s.Length;
            }

            if ( selStart >= 0 )
                textBox.SelectionStart = selStart;

            if ( selLength >= 0 )
                textBox.SelectionLength = selLength;
        }
        #endregion txtCode_MouseDoubleClick

        #region txtCode_KeyDown
        private void txtCode_KeyDown(object sender, KeyEventArgs e)
        {
            RichTextBox textBox = sender as RichTextBox;

            if ( e.Control && e.KeyCode == Keys.Space )
            {
                e.Handled = true;
                e.SuppressKeyPress = true;
                this.HandleStatementCompletion( textBox.SelectionStart, false );
            }
        }
        #endregion txtCode_KeyDown

        #region txtCode_KeyPress
        private void txtCode_KeyPress(object sender, KeyPressEventArgs e)
        {
            RichTextBox textBox = sender as RichTextBox;

            char c = e.KeyChar;
            bool isSpace = (int)c == 32;
            int selStart = textBox.SelectionStart;

            //  If the space key was pressed and the last token was an
            //  operator, show the statement completion dialog.
            if ( isSpace && selStart > 0 &&
                 this.IsOperator(textBox.Text[selStart - 1].ToString()) )
                this.HandleStatementCompletion( selStart + 1, true );
        }
        #endregion txtCode_KeyPress

        #region GetErrors
        private IEnumerable<IErrorToken> GetErrors( ImportResult result )
        {
            List<IErrorToken> errors = new List<IErrorToken>(result.Errors.Count);
            foreach( ImportError error in result.Errors )
            {
                errors.Add( error );
            }

            return errors;
        }

        #endregion GetErrors

        #region GetLineNumber
        protected override int GetLineNumber( string input, out int column )
        {
            return Grammar.GetLineNumber( input, this.importer.NewLine, out column );
        }
        #endregion GetLineNumber

        #region NewLine
        protected override string NewLine
        {
            get { return this.importer.NewLine; }
        }
        #endregion NewLine

        #region Grammar
        protected override Grammar Grammar { get { return this.importer.Grammar; } }
        #endregion Grammar

        #region TextHighlighter
        protected override TextHighlighter TextHighlighter { get { return this.highlighter; } }
        #endregion TextHighlighter

        #region InitializeOpenFileDialog / InitializeSaveFileDialog
        protected override void InitializeOpenFileDialog( OpenFileDialog dialog )
        {
            dialog.Filter = frmEBNFEditor.FILE_DIALOG_FILTER;
            dialog.FileName = string.Empty;
            dialog.InitialDirectory = CodeEditorBase.PATH_GRAMMAR_DEF;
        }

        protected override void InitializeSaveFileDialog( SaveFileDialog dialog )
        {
            dialog.Filter = frmEBNFEditor.FILE_DIALOG_FILTER;
            dialog.FileName = string.Empty;
            dialog.InitialDirectory = CodeEditorBase.PATH_GRAMMAR_DEF;
        }
        #endregion InitializeOpenFileDialog / InitializeSaveFileDialog

        #region HandleStatementCompletion
        private void HandleStatementCompletion( int selStart, bool auto )
        {
            Symbol declaration = null;
            StatementToken token = this.importer.FindToken( selStart, out declaration );

            if ( token == null )
                return;
            else
            if ( token != null )
            {
                //  If the current token is one for which we don't want to show the
                //  statement completion dialog, return
                if ( token is LiteralStringToken ||
                     token is SpecialSequence ||
                     token is CommentBlock )
                    return;

                //  If we are left of the assignment operator return
                if ( token is AssignmentToken && selStart <= token.CharacterPosition )
                    return;

                //  Handle literal strings differently
                TerminalSymbolToken tsToken = token as TerminalSymbolToken;
                if ( tsToken != null && tsToken.FromReference == false &&
                     declaration is TerminalSymbol )
                {
                    this.HandleStatementCompletion_TerminalSymbol( tsToken, tsToken.CharacterPosition, auto );
                    return;
                }

                //  If a NonTerminalSymbolDeclaration return
                NonTerminalSymbolDeclaration ntsDec = token as NonTerminalSymbolDeclaration;
                if ( ntsDec != null )
                    return;
            }

            bool findContext = (auto == false);

            Grammar grammar = this.Grammar;
            int start = selStart;
            string currentWord = findContext ? this.GetCurrentWord( selStart, out start ) : string.Empty;
            Symbol currentSymbol = null;
            Color tsColor = this.highlighter.ColorScheme.TerminalSymbolDeclarationForeColor;
            if ( tsColor.IsEmpty ){ tsColor = EBNFTokenColorScheme.DefaultTerminalSymbolDeclarationForeColor; }
            Color ntsColor = this.highlighter.ColorScheme.NonTerminalSymbolDeclarationForeColor;
            if ( ntsColor.IsEmpty ){ ntsColor = EBNFTokenColorScheme.DefaultNonTerminalSymbolDeclarationForeColor; }

            List<ImageListBox.Item> items = new List<ImageListBox.Item>( grammar.TerminalSymbols.Count + grammar.NonTerminalSymbols.Count );

            foreach ( TerminalSymbol ts in grammar.TerminalSymbols )
            {
                if ( ts.IsExplicitlyNamed == false )
                    continue;

                if ( findContext &&
                     currentSymbol == null &&
                     ts.Name.Equals(currentWord, StringComparison.InvariantCultureIgnoreCase) )
                    currentSymbol = ts;
 
                items.Add( new ImageListBox.Item(ts.Name, tsColor) );
            }

            foreach ( NonTerminalSymbol nts in grammar.NonTerminalSymbols )
            {
                //  Don't include the StartSymbol in the list because by doing
                //  so we would be encouraging a bad circular reference.
                if ( nts.IsStartSymbol )
                    continue;

                //  Also don't include the NonTerminalSymbol whose Rules collection
                //  the new symbol will belong, since that constitutes a bad circular
                //  reference as well.
                if ( declaration != null && declaration == nts )
                    continue;

                if ( findContext &&
                     currentSymbol == null &&
                     nts.Name.Equals(currentWord, StringComparison.InvariantCultureIgnoreCase) )
                    currentSymbol = nts;
 
                items.Add( new ImageListBox.Item(nts.Name, ntsColor) );
            }

            this.ShowStatementCompletionForm( start, items, currentWord, auto );
        }

        private void ShowStatementCompletionForm( int start, List<ImageListBox.Item> items, string currentWord, bool auto )
        {
            this.CreateStatementCompletionForm();

            Point location = this.txtCode.GetPositionFromCharIndex( start );
            location.Y += this.txtCode.Font.Height;
            location.X -= 16;
            location = this.txtCode.PointToScreen( location );

            this.statementCompletionForm.ShowMe(
                this,
                items,
                this.statementCompletionExclusions,
                currentWord,
                start,
                auto,
                location );
        }

        private void CreateStatementCompletionForm()
        {
            if ( this.statementCompletionForm == null )
            {
                if ( this.statementCompletionExclusions == null )
                    this.statementCompletionExclusions = this.GetStatementCompletionExclusions( this.importer.Notation );

                this.statementCompletionForm = new frmStatementCompletion();
                this.statementCompletionForm.Closed += new ClosedHandler(this.statementCompletionForm_Closed);
            }
        }

        private void HandleStatementCompletion_TerminalSymbol( TerminalSymbolToken tsToken, int selStart, bool auto )
        {
            List<ImageListBox.Item> items = new List<ImageListBox.Item>();

            Color color = this.highlighter.ColorScheme.SpecialSequenceSymbolForeColor;
            if ( color.IsEmpty ) { color = EBNFTokenColorScheme.DefaultSpecialSequenceSymbolForeColor; }
            string s = SpecialSequence.ToString( this.importer, SpecialSequenceType.TerminalSymbolLiteralIgnoreCase );
            items.Add( new ImageListBox.Item(s, color) );
            
            s = SpecialSequence.ToString( this.importer, SpecialSequenceType.TerminalSymbolRegex );
            items.Add( new ImageListBox.Item(s, color) );

            this.ShowStatementCompletionForm(
                selStart,
                items,
                null,
                auto );
        }

        private void statementCompletionForm_Closed(object sender, ClosedEventArgs e)
        {
            this.txtCode.Select( e.StartPos, e.CurrentText.Length );
            this.txtCode.SelectedText = e.Text;
        }

        #endregion HandleStatementCompletion

        #region IsOperator
        private bool IsOperator( string value )
        {
            if ( string.IsNullOrEmpty(value) )
                return false;

            GrammarNotation notation = this.importer.Notation;

            if ( value.Equals(notation.AlternationOperatorSymbol) ||
                 value.Equals(notation.ConcatenationOperatorSymbol) ||
                 value.Equals(notation.ConjunctionOperatorSymbol) ||
                 value.Equals(notation.NegationOperatorSymbol) ||
                 value.Equals(notation.RepetitionOperatorSymbol) ||
                 value.Equals(notation.GroupingBlockOpeningSymbol) ||
                 value.Equals(notation.OptionalBlockOpeningSymbol) ||
                 value.Equals(notation.RepetitionBlockOpeningSymbol) ||
                 value.Equals(notation.AssignmentSymbol) )
                return true;

            return false;
            
        }
        #endregion IsOperator

        #region GetStatementCompletionExclusions
        private List<string> GetStatementCompletionExclusions( GrammarNotation notation )
        {
            List<string> list = new List<string>();

            list.Add( notation.AlternationOperatorSymbol );
            list.Add( notation.AssignmentSymbol );
            list.Add( notation.CommentBlockClosingSymbol );
            list.Add( notation.CommentBlockOpeningSymbol );
            list.Add( notation.ConcatenationOperatorSymbol );
            list.Add( notation.ConjunctionOperatorSymbol );
            list.Add( notation.EndOfStatementSymbol );
            list.Add( notation.GroupingBlockClosingSymbol );
            list.Add( notation.GroupingBlockOpeningSymbol );
            list.Add( notation.NegationOperatorSymbol );
            list.Add( notation.OptionalBlockClosingSymbol );
            list.Add( notation.OptionalBlockOpeningSymbol );
            list.Add( notation.QuoteSymbol );
            list.Add( notation.RepetitionBlockClosingSymbol );
            list.Add( notation.RepetitionBlockOpeningSymbol );
            list.Add( notation.RepetitionOperatorSymbol );
            list.Add( notation.SpecialSequenceSymbol );

            return list;

        }
        #endregion GetStatementCompletionExclusions

    }
    #endregion frmEBNFEditor class
}
