﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace Infragistics.SyntaxParsing.Samples
{
    #region frmFind class
    public partial class frmFind : Form
    {
        private string text = null;
        private int currentPos = 0;

        public frmFind()
        {
            this.InitializeComponent();

            this.AcceptButton = this.cmdFindNext;
            this.KeyPreview = true;
            this.cmdFindNext.Click += new EventHandler(this.OnButtonClick);
            this.cmdReplace.Click += new EventHandler(this.OnButtonClick);
            this.cmdReplaceAll.Click += new EventHandler(this.OnButtonClick);
            this.txtFind.TextChanged += new EventHandler(this.OnTextBoxTextChanged);
            this.txtReplace.TextChanged += new EventHandler(this.OnTextBoxTextChanged);
        }

        #region OnTextBoxTextChanged
        private void OnTextBoxTextChanged(object sender, EventArgs e)
        {
            TextBox textBox = sender as TextBox;
            bool enabled = textBox.TextLength > 0;

            if ( sender == this.txtReplace )
            {
                this.cmdReplace.Enabled = enabled;
                this.cmdReplaceAll.Enabled = enabled;
            }
            else
            if ( sender == this.txtFind )
            {
                this.cmdFindNext.Enabled = enabled;
            }
        }
        #endregion OnTextBoxTextChanged

        public void ShowMe( Form owner, string text, string find, bool replace, int currentPos )
        {
            this.currentPos = currentPos;
            this.text = text;
            this.txtFind.Text = find;
            this.txtReplace.Text = null;
            this.txtFind.SelectAll();
            this.txtFind.Select();

            this.lblReplace.Enabled = replace;
            this.txtReplace.Enabled = replace;
            this.cmdReplace.Enabled = replace;
            this.cmdReplaceAll.Enabled = replace;

            if ( this.Visible == false )
                this.Show( owner );
        }

        public event FindHandler Find;
        public event ReplaceHandler Replace;
        public event ReplaceAllHandler ReplaceAll;

        private bool MatchCase { get { return this.chkMatchCase.Checked; } }

        private void OnButtonClick( object sender, EventArgs e )
        {
            if ( sender == this.cmdFindNext )
            {
                this._FindNext( this.currentPos, true );
                this.txtFind.SelectAll();
                this.txtFind.Select();
            }
            else
            if ( sender == this.cmdReplace )
            {
                this._Replace( this.currentPos );
                this.cmdReplace.Enabled = false;
            }
            else
            if ( sender == this.cmdReplaceAll )
            {
                this._ReplaceAll();
                this.cmdReplace.Enabled = false;
                this.cmdReplaceAll.Enabled = false;
            }
        }

        public bool FindNext()
        {
            int currentPos = this.currentPos;
            int index = this._FindNext( currentPos, true );
            return index != currentPos;
        }

        private int _FindNext( int startAt, bool fireEvent )
        {
            return this._FindNext( startAt, false, fireEvent );
        }

        private int _FindNext( int startAt, bool secondAttempt, bool fireEvent )
        {
            string find = this.txtFind.Text;
            if ( string.IsNullOrEmpty(find) )
                return -1;

            StringComparison comparison = this.MatchCase ? StringComparison.CurrentCulture : StringComparison.CurrentCultureIgnoreCase;
            string text = this.text.Substring( startAt );

            int index = text.IndexOf( find, comparison );
            FindEventArgs args = null;

            if ( index >= 0 )
            {
                int selStart = startAt + index;
                this.currentPos = selStart + find.Length;

                //  Success
                if ( fireEvent && this.Find != null )
                {
                    args = new FindEventArgs( find, selStart, true );
                    this.Find( this, args );
                    return selStart;
                }
                else
                    return selStart;
            }
            else
            if ( secondAttempt == false )
            {
                this.currentPos = 0;
                return this._FindNext( this.currentPos, true, fireEvent );
            }

            //  Failure
            if ( fireEvent && this.Find != null )
            {
                args = new FindEventArgs( find, -1, false );
                this.Find( this, args );
            }

            return -1;
        }

        private void _Replace( int startAt )
        {
            string find = this.txtFind.Text;
            if ( string.IsNullOrEmpty(find) )
                return;

            string replace = this.txtReplace.Text;
            if ( string.IsNullOrEmpty(replace) )
                return;

            int index = this.currentPos;
            string text = index >= 0 ? this.text.Substring( startAt, find.Length ) : string.Empty;

            if ( this.Replace != null )
            {
                ReplaceEventArgs args = new ReplaceEventArgs( find, replace, index );
                this.text.Remove(startAt, find.Length);
                this.text.Insert(startAt, replace);
                this.Replace( this, args );
            }
        }

        private void _ReplaceAll()
        {
            string find = this.txtFind.Text;
            if ( string.IsNullOrEmpty(find) )
                return;

            string replace = this.txtReplace.Text;
            if ( string.IsNullOrEmpty(replace) )
                return;

            this.text = this.text.Replace( find, replace );

            if ( this.ReplaceAll != null )
            {
                ReplaceEventArgsBase args = new ReplaceEventArgsBase( find, replace );
                this.ReplaceAll( this, args );
            }
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            base.OnClosing(e);

            if ( this.Disposing == false && this.IsDisposed == false )
            {
                this.Visible = false;
                e.Cancel = true;
            }
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            base.OnKeyDown(e);

            if ( e.KeyData == Keys.Escape )
                this.Close();
        }
    }
    #endregion frmFind class

    #region FindEventArgs class
    public delegate void FindHandler( object sender, FindEventArgs e );
    public class FindEventArgs : EventArgs
    {
        private bool success = false;
        private string text = null;
        private int charPos = 0;
        internal FindEventArgs( string text, int charPos, bool success )
        {
            this.text = text;
            this.charPos = charPos;
            this.success = success;
        }

        public string Text { get { return this.text; } }
        public int CharPos { get { return this.charPos; } }
        public bool Success { get { return this.success; } }
    }
    #endregion FindEventArgs class

    #region ReplaceEventArgs class
    public delegate void ReplaceHandler( object sender, ReplaceEventArgs e );
    public delegate void ReplaceAllHandler( object sender, ReplaceEventArgsBase e );
    public class ReplaceEventArgsBase : EventArgs
    {
        private string findText = null;
        private string replaceText = null;

        internal ReplaceEventArgsBase( string findText, string replaceText )
        {
            this.findText = findText;
            this.replaceText = replaceText;
        }

        public string FindText { get { return this.findText; } }
        public string ReplaceText { get { return this.replaceText; } }
    }

    public class ReplaceEventArgs : ReplaceEventArgsBase
    {
        private int charPos = 0;
        internal ReplaceEventArgs( string findText, string replaceText, int charPos ) : base( findText, replaceText )
        {
            this.charPos = charPos;
        }

        public int CharPos { get { return this.charPos; } }
    }

    #endregion ReplaceEventArgs class
}
