﻿#region Copyright
//
//  The token highlighting functionality for this class was based
//  on the following codeproject article:
//-----------------------------------------------------------------------
//
//  http://www.codeproject.com/KB/recipes/TinyPG.aspx
//
// Copyright 2008 - 2010 Herre Kuijpers - <herre.kuijpers@gmail.com>
//
// This source file(s) may be redistributed, altered and customized
// by any means PROVIDING the authors name and all copyright
// notices remain intact.
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED. USE IT AT YOUR OWN RISK. THE AUTHOR ACCEPTS NO
// LIABILITY FOR ANY DATA DAMAGE/LOSS THAT THIS PRODUCT MAY CAUSE.
//
//  SEE ALSO:
//  Parsing Expression Grammar Support for C# 3.0 Part 1 - PEG Lib and Parser:
//  http://www.codeproject.com/KB/recipes/grammar_support_1.aspx
//
//-----------------------------------------------------------------------
#endregion Copyright

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Threading;
using System.Diagnostics;

using Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;

namespace Infragistics.SyntaxParsing.Samples
{
    #region EBNFTextHighlighter
    /// <summary>
    /// Highlights text in an EBNF script
    /// </summary>
    public class EBNFTextHighlighter : TextHighlighter
    {
        private GrammarImporter importer = null;
        private ImportResult lastResult = null;
        private List<StatementToken> visualTokens = null;
        private EBNFTokenColorScheme colorScheme = null;

        internal EBNFTextHighlighter( RichTextBox textBox, GrammarImporter importer ) : this( textBox, importer, null )
        {
        }

        internal EBNFTextHighlighter( RichTextBox textBox, GrammarImporter importer, EBNFTokenColorScheme colorScheme ) : base( textBox )
        {
            if ( importer == null )
                throw new ArgumentNullException("textBox");

            this.importer = importer;
            this.colorScheme = colorScheme != null ? colorScheme : new EBNFTokenColorScheme( importer );
            this.colorScheme.InitializeImporter( this.importer );
        }

        internal override void OnTextBoxSelectionChanged( int selectionStart )
        {
        }

        protected override void AssignParseResult( object value )
        {
            this.visualTokens = value as List<StatementToken>;
        }

        protected override void HighlightTextCore()
        {
            StringBuilder sb = new StringBuilder();
            this.HighlightTokens( this.visualTokens, sb );
            this.AddRtfHeader( sb );
            this.AddRtfEnd( sb );
            this.VerifyEncoding( sb );
            this.textBox.Rtf = sb.ToString();

            if ( this.lastResult != null && this.CompileComplete != null )
                this.CompileComplete.Invoke( this, new CompileCompleteEventArgs(this.lastResult) );
        }

        protected override object Parse(string text)
        {
            this.lastResult = this.importer.Load( text );
            return importer.Tokens;
        }

        private void HighlightTokens( List<StatementToken> tokens, StringBuilder sb )
        {
            foreach ( StatementToken token in tokens )
            {
                int id = this.colorScheme.GetColorId( token );

                //  BF 9/13/11 WEIRD
                //if ( id > 0 )
                if ( id >= 0 )
                    sb.Append( string.Format( @"{{\cf{0} ", id) );

                string tokenText = token.Text;

                TerminalSymbolToken tsToken = token as TerminalSymbolToken;
                NonTerminalSymbolToken ntsToken = tsToken == null ? token as NonTerminalSymbolToken : null;

                if ( tsToken != null && tsToken.FromReference )
                    tokenText = tsToken.TerminalSymbol.Name;
                else
                if ( ntsToken != null )
                    tokenText = ntsToken.NonTerminalSymbol.Name;

                tokenText = TextHighlighter.EscapeForRtf( tokenText );

                sb.Append( tokenText );

                if ( id > 0 )
                    sb.Append( "}" );
            }
        }

        protected override void AddRtfHeader( StringBuilder sb )
        {
            //sb.Insert(0, @"{\rtf1\ansi\deff0{\fonttbl{\f0\fnil\fcharset0 Consolas;}}{\colortbl;\red0\green128\blue0;\red0\green128\blue0;\red255\green0\blue0;\red128\green0\blue255;\red128\green0\blue128;\red128\green0\blue128;\red43\green145\blue202;\red0\green0\blue255;\red255\green0\blue0;\red0\green0\blue255;\red43\green145\blue202;\red0\green128\blue0;\red0\green128\blue0;\red163\green21\blue21;\red0\green128\blue0;\red0\green128\blue0;\red163\green21\blue21;\red0\green128\blue0;\red0\green128\blue0;\red163\green21\blue21;\red128\green0\blue128;\red128\green0\blue128;\red0\green0\blue255;\red128\green0\blue128;\red163\green21\blue21;}\viewkind4\uc1\pard\lang1033\f0\fs20");

            string fontName = this.textBox != null ? this.textBox.Font.Name : "Lucida Console";
            string colors = this.colorScheme.SerializedColorTable;
            //string header = string.Format( TextHighlighter.RtfHeaderTemplate, fontName, colors );
            string header = TextHighlighter.RtfHeaderTemplate.Replace( "{0}", fontName );
            header = header.Replace( "{1}", colors );
            sb.Insert(0, header);
        }

        protected override void OnMessagesResumed()
        {
            if ( this.lastResult != null && this.lastResult.Success == false )
            {
                for ( int i = 0; i < this.lastResult.Errors.Count; i ++ )
                {
                    ImportError error = this.lastResult.Errors[i];
                    int len = error.Text.Length > 0 ? error.Text.Length : 1;
                    this.Underline( error.CharacterPosition, len, false );
                }
            }
        }

        public event CompileCompleteHandler CompileComplete;

        public EBNFTokenColorScheme ColorScheme { get { return this.colorScheme; } }
    }
    #endregion EBNFTextHighlighter

    #region EBNFTokenColorScheme
    public class EBNFTokenColorScheme
    {
        #region Constants
        public static readonly Color DefaultAssignmentSymbolForeColor = Color.Empty;
        public static readonly Color DefaultEndOfStatementSymbolForeColor = Color.Empty;
        public static readonly Color DefaultAlternationOperatorForeColor = Color.Empty;
        public static readonly Color DefaultConcatenationOperatorForeColor = Color.Empty;
        public static readonly Color DefaultConjunctionOperatorForeColor = Color.Empty;
        public static readonly Color DefaultNegationOperatorForeColor = Color.Empty;
        public static readonly Color DefaultRepetitionOperatorForeColor = Color.Empty;
        public static readonly Color DefaultRepetitionBlockSymbolForeColor = Color.Empty;
        public static readonly Color DefaultOptionalBlockSymbolForeColor = Color.Empty;
        public static readonly Color DefaultGroupingBlockSymbolForeColor = Color.Empty;
        public static readonly Color DefaultSpecialSequenceSymbolForeColor = Color.DarkGray;
        public static readonly Color DefaultCommentBlockForeColor = Color.DarkGreen;
        public static readonly Color DefaultTerminalSymbolDeclarationForeColor = Color.Magenta;
        public static readonly Color DefaultNonTerminalSymbolDeclarationForeColor = Color.Blue;
        public static readonly Color DefaultTerminalSymbolForeColor = Color.Firebrick;
        public static readonly Color DefaultNonTerminalSymbolForeColor = Color.FromArgb( 85, 145, 189 );
        public static readonly Color DefaultLiteralStringForeColor = EBNFTokenColorScheme.DefaultTerminalSymbolForeColor;
        #endregion Constants

        #region Member variables
        private GrammarImporter importer = null;
        private Dictionary<StatementTokenType, int>   tokenTable = null;
        private Dictionary<int, Color>   colorTable = null;
        private string serializedColorTable = null;
        #endregion Member variables

        #region Constructor
        public EBNFTokenColorScheme( GrammarImporter importer )
        {
            this.InitializeImporter( importer );
        }

        internal void InitializeImporter( GrammarImporter importer ){ this.importer = importer; }

        #endregion Constructor

        #region Properties
        public Color AssignmentSymbolForeColor { get; set; }
        public Color EndOfStatementSymbolForeColor { get; set; }
        public Color AlternationOperatorForeColor { get; set; }
        public Color ConcatenationOperatorForeColor { get; set; }
        public Color ConjunctionOperatorForeColor { get; set; }
        public Color NegationOperatorForeColor { get; set; }
        public Color RepetitionOperatorForeColor { get; set; }
        public Color RepetitionBlockOpeningSymbolForeColor { get; set; }
        public Color RepetitionBlockSymbolForeColor { get; set; }
        public Color OptionalBlockSymbolForeColor { get; set; }
        public Color GroupingBlockSymbolForeColor { get; set; }
        public Color SpecialSequenceSymbolForeColor { get; set; }
        public Color CommentBlockForeColor { get; set; }
        public Color TerminalSymbolDeclarationForeColor { get; set; }
        public Color NonTerminalSymbolDeclarationForeColor { get; set; }
        public Color TerminalSymbolForeColor { get; set; }
        public Color NonTerminalSymbolForeColor { get; set; }
        public Color LiteralStringForeColor { get; set; }
        #endregion Properties

        #region Methods
        
        public int GetColorId( StatementToken token )
        {
            if ( token == null )
                return 0;

            this.BuildColorTable();

            GrammarNotation notation = this.importer.Notation;
            StatementTokenType tokenType = token.TokenType;
            if ( tokenType == StatementTokenType.BlockSymbol )
            {                   
                BlockSymbol blockSymbol = token as BlockSymbol;

                if ( blockSymbol.Text.Equals(notation.GroupingBlockOpeningSymbol) ||
                     blockSymbol.Text.Equals(notation.GroupingBlockClosingSymbol) )
                {
                    tokenType = StatementTokenType.Group;
                }
                else
                if ( blockSymbol.Text.Equals(notation.RepetitionBlockOpeningSymbol) ||
                     blockSymbol.Text.Equals(notation.RepetitionBlockClosingSymbol) )
                {
                    tokenType = StatementTokenType.RepetitionGroup;
                }
                else
                if ( blockSymbol.Text.Equals(notation.OptionalBlockOpeningSymbol) ||
                     blockSymbol.Text.Equals(notation.OptionalBlockClosingSymbol) )
                {
                    tokenType = StatementTokenType.OptionalGroup;
                }
            }

            int id = 0;
            if ( this.tokenTable.TryGetValue(tokenType, out id) )
                return id;
            else
                return 0;
        }

        private void BuildColorTable()
        {
            if ( this.tokenTable == null )
                this.tokenTable = new Dictionary<StatementTokenType,int>();
            else
                return;

            if ( this.colorTable == null )
                this.colorTable = new Dictionary<int,Color>();
            else
                this.colorTable.Clear();

            Color color = Color.Empty;

            List<ColorPair> pairs = new List<ColorPair>();

            pairs.Add( new ColorPair(StatementTokenType.Assignment, this.AssignmentSymbolForeColor, DefaultAssignmentSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.EndOfStatement, this.EndOfStatementSymbolForeColor, DefaultEndOfStatementSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.AlternationOperator, this.AlternationOperatorForeColor, DefaultAlternationOperatorForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.ConcatenationOperator, this.ConcatenationOperatorForeColor, DefaultConcatenationOperatorForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.ConjunctionOperator, this.ConjunctionOperatorForeColor, DefaultConjunctionOperatorForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.NegationOperator, this.NegationOperatorForeColor, DefaultNegationOperatorForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.SyntacticFactor, this.RepetitionOperatorForeColor, DefaultRepetitionOperatorForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.RepetitionGroup, this.RepetitionBlockSymbolForeColor, DefaultRepetitionBlockSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.OptionalGroup, this.OptionalBlockSymbolForeColor, DefaultOptionalBlockSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.Group, this.GroupingBlockSymbolForeColor, DefaultGroupingBlockSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.SpecialSequence, this.SpecialSequenceSymbolForeColor, DefaultSpecialSequenceSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.CommentBlock, this.CommentBlockForeColor, DefaultCommentBlockForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.TerminalSymbolDeclaration, this.TerminalSymbolDeclarationForeColor, DefaultTerminalSymbolDeclarationForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.NonTerminalSymbolDeclaration, this.NonTerminalSymbolDeclarationForeColor, DefaultNonTerminalSymbolDeclarationForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.TerminalSymbol, this.TerminalSymbolForeColor, DefaultTerminalSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.NonTerminalSymbol, this.NonTerminalSymbolForeColor, DefaultNonTerminalSymbolForeColor) );
            pairs.Add( new ColorPair(StatementTokenType.LiteralString, this.LiteralStringForeColor, DefaultLiteralStringForeColor) );

            int id = 1;
            foreach ( ColorPair pair in pairs )
            {
                color = pair.Color1.IsEmpty == false ? pair.Color1 : pair.Color2;

                if ( color.IsEmpty == false )
                {
                    this.tokenTable.Add(pair.TokenType, id );
                    this.colorTable.Add( id, color );
                    id += 1;
                }
            }
        }

        public string SerializedColorTable
        {
            get
            {
                if ( string.IsNullOrEmpty(this.serializedColorTable) )
                    this.serializedColorTable = this.SerializeColorTable();

                return this.serializedColorTable;
            }
        }

        private string SerializeColorTable()
        {
            this.BuildColorTable();

            return
                this.colorTable != null && this.colorTable.Count > 0 ?
                TextHighlighter.SerializeColors( this.colorTable.Values ) :
                string.Empty;
        }

        #endregion Methods

        #region ColorPair class
        private class ColorPair
        {
            public ColorPair ( StatementTokenType tokenType, Color color1, Color color2 )
            {
                this.TokenType = tokenType;
                this.Color1 = color1;
                this.Color2 = color2;
            }

            public StatementTokenType TokenType { get; set; }
            public Color Color1 { get; set; }
            public Color Color2 { get; set; }

            public override string ToString()
            {
                return this.TokenType.ToString();
            }
        }
        #endregion ColorPair class
    }
    #endregion EBNFTokenColorScheme

    #region CompileCompleteEventArgs
    
    public delegate void CompileCompleteHandler( object sender, CompileCompleteEventArgs e );

    public class CompileCompleteEventArgs : EventArgs
    {
        private ImportResult result = null;

        public CompileCompleteEventArgs( ImportResult result )
        {
            this.result = result;
        }

        public ImportResult Result { get { return this.result; } }
    }

    #endregion CompileCompleteEventArgs
}
