﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.IO;

using Infragistics.SyntaxParsing;
using IGGrammar = Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;

namespace Infragistics.SyntaxParsing.Samples
{
    #region ErrorTokenToolTip class
    /// <summary>
    /// Maybe it's just me but the Windows Forms
    /// ToolTip was excruciatingly disfunctional
    /// so I did this to work around it.
    /// </summary>
    public class ErrorTokenToolTip : ToolTip
    {
        private const int TIMER_INTERVAL = 100;
        private bool ignoreNextMouseMove = false;
        private bool toolTipVisible = false;
        private RichTextBox control = null;
        private Timer timer = null;
        private int lastMouseMove = 0;
        private IEnumerable<IErrorToken> errors = null;

        internal ErrorTokenToolTip( RichTextBox control )
        {
            this.UseAnimation = this.UseFading = false;
            this.InitialDelay = 0;
            this.control = control;
            this.control.MouseMove += new MouseEventHandler(this.control_MouseMove);
            this.timer = new Timer();
            this.timer.Interval = ErrorTokenToolTip.TIMER_INTERVAL;
            this.timer.Tick += new EventHandler(this.OnTimerTick);
        }

        public void Enable( IEnumerable<IErrorToken> errors )
        {
            this.errors = errors;
            this.timer.Enabled = this.errors != null;
        }

        #region control_MouseMove
        private void control_MouseMove(object sender, MouseEventArgs e)
        {
            if ( this.ignoreNextMouseMove )
            {
                this.ignoreNextMouseMove = false;
                return;
            }

            this.Hide(this.control);
            this.toolTipVisible = false;
            this.lastMouseMove = Environment.TickCount;
        }
        #endregion control_MouseMove

        #region OnTimerTick
        private void OnTimerTick( object sender, EventArgs e )
        {
            if ( this.toolTipVisible || this.errors == null ||
                 this.control.IsDisposed || this.control.Disposing )
                return;

            Timer timer = sender as Timer;
            int ticks = Environment.TickCount - this.lastMouseMove;
            if ( ticks > timer.Interval )
            {
                Rectangle bounds = this.control.Bounds;
                Point cursorPos = this.control.PointToClient(Cursor.Position);
                if ( cursorPos.X >= 0 && cursorPos.X < bounds.Width &&
                     cursorPos.Y >= 0 && cursorPos.Y < bounds.Height )
                {                                    
                    int charPos = this.control.GetCharIndexFromPosition( cursorPos );

                    //  It was either this or let the tooltip show all the time
                    //  because GetCharIndexFromPosition returns a valid index
                    //  even when the cursor is miles away from the last char.
                    if ( charPos < 0 || charPos >= (this.control.TextLength - 1) )
                        return;

                    foreach( IErrorToken error in this.errors )
                    {
                        if ( charPos >= error.CharacterPosition &&
                             charPos <= (error.CharacterPosition + error.Text.Length) )
                        {
                            Point showAt = this.control.GetPositionFromCharIndex( error.CharacterPosition );
                            showAt.Y += (int)(Math.Ceiling(this.control.Font.Size * 2));

                            this.Show( error.Description, this.control, showAt );
                            this.ignoreNextMouseMove = true;
                            this.toolTipVisible = true;
                        }
                    }
                }

            }
        }
        #endregion OnTimerTick

        #region Dispose
        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);

            if ( disposing )
            {
                if ( this.timer != null )
                {
                    this.timer.Tick -= new EventHandler(this.OnTimerTick);
                    this.timer.Stop();
                    this.timer.Dispose();
                    this.timer = null;
                }

                if ( this.control != null )
                {
                    this.control.MouseMove -= new MouseEventHandler(this.control_MouseMove);
                    this.control = null;
                }
            }
        }
        #endregion Dispose

    }
    #endregion ErrorTokenToolTip class
}
