#region Copyright
//
//  The token highlighting functionality for this class was based
//  on the following codeproject article:
//-----------------------------------------------------------------------
//
//  http://www.codeproject.com/KB/recipes/TinyPG.aspx
//
// Copyright 2008 - 2010 Herre Kuijpers - <herre.kuijpers@gmail.com>
//
// This source file(s) may be redistributed, altered and customized
// by any means PROVIDING the authors name and all copyright
// notices remain intact.
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED. USE IT AT YOUR OWN RISK. THE AUTHOR ACCEPTS NO
// LIABILITY FOR ANY DATA DAMAGE/LOSS THAT THIS PRODUCT MAY CAUSE.
//
//  SEE ALSO:
//  Parsing Expression Grammar Support for C# 3.0 Part 1 - PEG Lib and Parser:
//  http://www.codeproject.com/KB/recipes/grammar_support_1.aspx
//
//-----------------------------------------------------------------------
#endregion Copyright

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Threading;
using System.Diagnostics;

using Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;

namespace Infragistics.SyntaxParsing.Samples
{
    #region TextHighlighter
    /// <summary>
    /// Takes control over the RichTextBox and will color the text accoording to the rules of the parser and the scanner
    /// this control extender will also support Undo/Redo functionality.
    /// </summary>
    public abstract class TextHighlighter : IDisposable
    {
        #region API declarations
        
        [DllImport("user32", CharSet = CharSet.Auto)]
        private extern static IntPtr SendMessage(IntPtr hWnd, int msg, int wParam, IntPtr lParam);

        [DllImport("user32", CharSet = CharSet.Auto)]
        private static extern int SendMessage(HandleRef hWnd, int msg, int wParam, ref CHARFORMAT lp);

        [DllImport("user32.dll")]
        private static extern bool PostMessageA(IntPtr hWnd, int nBar, int wParam, int lParam);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        private static extern int GetScrollPos(int hWnd, int nBar);

        [DllImport("user32.dll")]
        private static extern int SetScrollPos(IntPtr hWnd, int nBar, int nPos, bool bRedraw);
        
        #endregion API declarations

        #region Constants

        private const int WM_SETREDRAW = 0x000B;
        private const int WM_USER = 0x400;
        private const int EM_GETEVENTMASK = (WM_USER + 59);
        private const int EM_SETEVENTMASK = (WM_USER + 69);
        private const int SB_HORZ = 0x0;
        private const int SB_VERT = 0x1;
        private const int WM_HSCROLL = 0x114;
        private const int WM_VSCROLL = 0x115;
        private const int SB_THUMBPOSITION = 4;
        private const int UNDO_BUFFER = 100;

  
        //  Underlining
        private const int SCF_SELECTION = 1;
        private const int CFM_UNDERLINETYPE = 8388608;
        private const int CFE_UNDERLINE = 4;
        private const int EM_SETCHARFORMAT = 1092;
        private const int EM_GETCHARFORMAT = 1082;
        private const int UNDERLINE_COLOR_RED = 0x50;

        /// <summary>
        /// {0} = font name
        /// {1} = color table entries
        /// </summary>
        internal const string RtfHeaderTemplate = @"{\rtf1\ansi\deff0{\fonttbl{\f0\fnil\fcharset0 {0};}}{\colortbl;{1}}\viewkind4\uc1\pard\lang1033\f0\fs20";

        static private Encoding TextEncoding = frmMain.TextEncoding;

        private const int REFRESH_INTERVAL = 500;
        
        #endregion Constants

        #region Member variables
        
        protected RichTextBox textBox;
        public static object lockObject = new object();

        protected IntPtr stateLocked = IntPtr.Zero;

        private Thread threadAutoHighlight;

        private bool isDisposing;
        private bool textChanged;
        private string currentText;

        public event EventHandler Highlighting;
        public event EventHandler HighlightComplete;

        System.Threading.Timer timer = null;
        private bool suspended = false;
        
        #endregion Member variables

        #region HScrollPos / VScrollPos
        private int HScrollPos
        {
            get { return GetScrollPos((int)textBox.Handle, SB_HORZ); }
            set
            {
                SetScrollPos((IntPtr)textBox.Handle, SB_HORZ, value, true);
                PostMessageA((IntPtr)textBox.Handle, WM_HSCROLL, SB_THUMBPOSITION + 0x10000 * value, 0);
            }
        }

        private int VScrollPos
        {
            get { return GetScrollPos((int)textBox.Handle, SB_VERT); }
            set
            {
                SetScrollPos((IntPtr)textBox.Handle, SB_VERT, value, true);
                PostMessageA((IntPtr)textBox.Handle, WM_VSCROLL, SB_THUMBPOSITION + 0x10000 * value, 0);
            }
        }
        #endregion HScrollPos / VScrollPos

        #region Constructor
        protected TextHighlighter( RichTextBox textbox )
        {
            this.textBox = textbox;

            TimerCallback callback = new TimerCallback( this.OnTimerTick );
            this.timer = new System.Threading.Timer( callback, null, Timeout.Infinite, frmMain.REFRESH_TIMER_INTERVAL );

            this.textBox.TextChanged += new EventHandler(Textbox_TextChanged);
            this.textBox.SelectionChanged += new EventHandler(Textbox_SelectionChanged);
            this.textBox.Disposed += new EventHandler(Textbox_Disposed);

            this.threadAutoHighlight = new Thread(this.AutoHighlightStart);
            this.threadAutoHighlight.Start();
        }
        #endregion Constructor

        #region Lock / Unlock
        public void Lock()
        {
            // Suspend painting operations  
            SendMessage(textBox.Handle, WM_SETREDRAW, 0, IntPtr.Zero);

            // Suspend messages
            stateLocked = SendMessage(textBox.Handle, EM_GETEVENTMASK, 0, IntPtr.Zero);
        }

        public void Unlock()
        {
            // Resume messages 
            SendMessage(textBox.Handle, EM_SETEVENTMASK, 0, stateLocked);

            this.OnMessagesResumed();

            // Resume painting
            SendMessage(textBox.Handle, WM_SETREDRAW, 1, IntPtr.Zero);

            stateLocked = IntPtr.Zero;

            //  Invalidate
            textBox.Invalidate();
        }
        #endregion Lock / Unlock

        #region OnMessagesResumed
        protected virtual void OnMessagesResumed(){}
        #endregion OnMessagesResumed

        #region Suspend/Resume
        public void Suspend() { this.suspended = true; }
        public void Resume() { this.suspended = false; }
        #endregion Suspend/Resume

        #region RichTextBox events

        private void Textbox_TextChanged(object sender, EventArgs e)
        {
            if ( stateLocked != IntPtr.Zero )
                return;

            if ( this.suspended )
                this.HighlightText();
            else
                this.timer.Change( frmMain.REFRESH_TIMER_INTERVAL, frmMain.REFRESH_TIMER_INTERVAL );
        }

        private void Textbox_SelectionChanged(object sender, EventArgs e)
        {
            this.OnTextBoxSelectionChanged( this.textBox.SelectionStart );            
        }

        internal abstract void OnTextBoxSelectionChanged( int selectionStart );

        private void Textbox_Disposed(object sender, EventArgs e)
        {
            this.Dispose();
        }

        #endregion RichTextBox events

        #region HighlightText

        /// <summary>
        /// use HighlighText to start the text highlight process from the caller's thread.
        /// this method is not used internally. 
        /// </summary>
        public void HighlightText()
        {
            lock ( lockObject )
            {
                this.textChanged = true;
                this.currentText = this.textBox.Text;
            }
        }


        // highlight the text (used internally only)
        private void HighlightTextInternal()
        {
            if ( this.Highlighting != null )
                this.Highlighting.Invoke( this, EventArgs.Empty );

            this.Lock();

            int hscroll = this.HScrollPos;
            int vscroll = this.VScrollPos;

            int selStart = this.textBox.SelectionStart;
            int selLength = this.textBox.SelectionLength;

            this.HighlightTextCore();

            this.textBox.Select(selStart, selLength);

            this.HScrollPos = hscroll;
            this.VScrollPos = vscroll;

            this.Unlock();

            if ( this.HighlightComplete != null )
                this.HighlightComplete.Invoke( this, EventArgs.Empty );
        }

        protected abstract void HighlightTextCore();

        #endregion HighlightText

        #region Underline
        internal void Underline( int selStart, int selLength, bool clear )
        {
            if ( selStart >= this.textBox.Text.Length )
                return;

            int color = clear ? 0 : TextHighlighter.UNDERLINE_COLOR_RED;
            int style = clear ? 0 : 8;

            CHARFORMAT fmt = new CHARFORMAT();
            fmt.cbSize = Marshal.SizeOf(fmt);
            fmt.dwMask = CFM_UNDERLINETYPE;
            if ( clear == false )
            {
                fmt.dwEffects = CFE_UNDERLINE;
            }
            
            fmt.bUnderlineType = (byte)((byte)style | (byte)color);

            int _selStart = this.textBox.SelectionStart;

            //  Set selection
            this.textBox.SelectionStart = selStart;
            this.textBox.SelectionLength = selLength;

            // Set the underline color
            HandleRef handleRef = new HandleRef(this.textBox, this.textBox.Handle);
            SendMessage( handleRef, EM_SETCHARFORMAT, SCF_SELECTION, ref fmt);

            //  Restore selection
            this.textBox.SelectionStart = _selStart;
            this.textBox.SelectionLength = 0;
        }
        #endregion Underline

        #region Unicode
        /// <summary>
        /// Converts each character to the unicode equivalent.
        /// </summary>
        public StringBuilder Unicode(StringBuilder sb)
        {
            int i = 0;
            StringBuilder uc = new StringBuilder();
            for (i = 0; i <= sb.Length - 1; i++)
            {
                char c = sb[i];
                
                if ((int)c < 127)
                {
                    uc.Append(c);
                }
                else
                {
                    uc.Append("\\u");
                    uc.Append(((int)c).ToString());
                    uc.Append("?");
                }
            }
            return uc;
        }
        #endregion Unicode

        #region AutoHighlightStart
        private void AutoHighlightStart()
        {
            string _currenttext = string.Empty;
            while ( this.isDisposing == false )
            {
                bool _textchanged;

                lock (lockObject)
                {
                    _textchanged = this.textChanged;
                    if ( textChanged )
                    {
                        this.textChanged = false;
                        _currenttext = currentText;
                    }
                }

                if ( _textchanged == false )
                {
                    Thread.Sleep(TextHighlighter.REFRESH_INTERVAL);
                    continue;
                }

                object parseResult = this.Parse( _currenttext );

                lock ( lockObject )
                {
                    if ( this.textChanged )
                        continue;
                    else
                        this.AssignParseResult( parseResult );
                }
                
                this.textBox.Invoke(new MethodInvoker(this.HighlightTextInternal));
            }
        }

        protected abstract object Parse( string text );

        protected abstract void AssignParseResult( object value );

        #endregion AutoHighlightStart

        #region Debug
        //internal void DebugWriteLine()
        //{
        //    Debug.WriteLine( this.textBox.Rtf );
        //}

        //internal void DebugWriteString( string s )
        //{
        //    Debug.WriteLine( s );
        //}
        #endregion Debug

        #region AddRtfHeader
        // define the color palette to be used here
        protected virtual void AddRtfHeader(StringBuilder sb)
        {
            //sb.Insert(0, @"{\rtf1\ansi\deff0{\fonttbl{\f0\fnil\fcharset0 Consolas;}}{\colortbl;\red0\green128\blue0;\red0\green128\blue0;\red255\green0\blue0;\red128\green0\blue255;\red128\green0\blue128;\red128\green0\blue128;\red43\green145\blue202;\red0\green0\blue255;\red255\green0\blue0;\red0\green0\blue255;\red43\green145\blue202;\red0\green128\blue0;\red0\green128\blue0;\red163\green21\blue21;\red0\green128\blue0;\red0\green128\blue0;\red163\green21\blue21;\red0\green128\blue0;\red0\green128\blue0;\red163\green21\blue21;\red128\green0\blue128;\red128\green0\blue128;\red0\green0\blue255;\red128\green0\blue128;\red163\green21\blue21;}\viewkind4\uc1\pard\lang1033\f0\fs20");
            sb.Insert(0, @"{\rtf1\ansi\deff0{\fonttbl{\f0\fnil\fcharset0 Consolas;}}{\colortbl;\red255\green0\blue0;\red0\green128\blue0;\red0\green0\blue255;}\viewkind4\uc1\pard\lang1033\f0\fs20");
        }
        #endregion AddRtfHeader

        #region AddRtfEnd
        protected void AddRtfEnd(StringBuilder sb)
        {
            sb.Append("}");
        }
        #endregion AddRtfEnd

        #region SerializeColors
        static internal string SerializeColors(IEnumerable<Color> colors)
        {
            if ( colors == null )
                return string.Empty;

            StringBuilder sb = new StringBuilder();

            foreach( Color color in colors )
            {
                sb.Append( TextHighlighter.ColorToRtf(color) );
            }

            return sb.ToString();
        }
        #endregion SerializeColors

        #region VerifyEncoding
        /// <summary>
        /// This doesn't work, and I don't know why.
        /// </summary>
        protected StringBuilder VerifyEncoding( StringBuilder sb )
        {
            if ( TextHighlighter.TextEncoding == Encoding.ASCII )
                return sb;
            else
                return this.Unicode( sb );
        }
        #endregion VerifyEncoding

        #region OnTimerTick
        private void OnTimerTick( object state )
        {
            this.textBox.Invoke( new MethodInvoker(this.HighlightText) );
            this.timer.Change( Timeout.Infinite, frmMain.REFRESH_TIMER_INTERVAL );
        }
        #endregion OnTimerTick

        #region UndoItem class
        private class UndoItem
        {
            /// <summary>
            /// contains the information for an undo/redo action
            /// </summary>
            /// <param name="text">the full text to be undone/redone</param>
            /// <param name="position">position of the caret after the un/redo action</param>
            /// <param name="scroll">position of the scrollbars after un/redo action</param>
            public UndoItem(string text, int position, Point scroll)
            {
                Text = text;
                Position = position;
                ScrollPosition = scroll;
            }

            public string Text;
            public int Position;
            public Point ScrollPosition;
        }
        #endregion UndoItem class

        #region IDisposable Members

        void IDisposable.Dispose()
        {
            this.Dispose();
        }

        public void Dispose()
        {
            this.isDisposing = true;

            this.threadAutoHighlight.Join(1000);

            if ( this.threadAutoHighlight.IsAlive )
                this.threadAutoHighlight.Abort();
        }

        #endregion IDisposable Members

        #region ColorToRtf
        static internal string ColorToRtf( Color color )
        {
            if ( color.IsEmpty )
                return string.Empty;

            return string.Format( "\\red{0}\\green{1}\\blue{2};", color.R, color.G, color.B );
        }
        #endregion ColorToRtf

        #region EscapeForRtf
        static internal string EscapeForRtf( string value )
        {
            if ( string.IsNullOrEmpty(value) )
                return string.Empty;

            value = value.Replace(@"\", @"\\");
            value = value.Replace("{", @"\{");
            value = value.Replace("}", @"\}");
            value = value.Replace("\r\n", "\\par\n");
            value = value.Replace("\n", "\\par\n");

            return value;
        }
        #endregion EscapeForRtf

        #region CHARFORMAT struct
        [StructLayout(LayoutKind.Sequential)]
        private struct CHARFORMAT
        {
            public int cbSize;
            public uint dwMask;
            public uint dwEffects;
            public int yHeight;
            public int yOffset;
            public int crTextColor;
            public byte bCharSet;
            public byte bPitchAndFamily;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 32)]
            public char[] szFaceName;

            public short wWeight;
            public short sSpacing;
            public int crBackColor;
            public int LCID;
            public uint dwReserved;
            public short sStyle;
            public short wKerning;
            public byte bUnderlineType;
            public byte bAnimation;
            public byte bRevAuthor;
        }
        #endregion CHARFORMAT struct
    }
    #endregion TextHighlighter
}
