﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;

using Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;

namespace Infragistics.SyntaxParsing.Samples
{
    #region frmMain class
    public partial class frmMain : CodeEditorBase
    {
        #region Constants
        private const string APP_NAME = "Infragistics code editor";
        private const string FILE_DIALOG_FILTER = "All files (*.*)|*.*";
        #endregion Constants

        #region Member variables

        private Grammar grammar = new Grammar();
        private string scriptFilename = null;

        private ParseTree parseTree = null;

        private ParseTreeTextHighlighter highlighter = null;

        private Regex regexTabs = new Regex( "[\t]+" );
        private Regex regexLine = new Regex( ".*[\n]" );
        private Regex regexBackToNewline = new Regex( "[\n]", RegexOptions.RightToLeft );

        private int compileStart = 0;

        #endregion Member variables

        #region Constructor

        public frmMain()
        {
            this.InitializeComponent();
            this.InitializeUI();
        }

        #endregion Constructor

        #region Grammar
        protected override Grammar Grammar { get { return this.grammar; } }
        #endregion Grammar
                
        #region TextHighlighter
        protected override TextHighlighter TextHighlighter { get { return this.highlighter; } }
        #endregion TextHighlighter

        #region InitializeUI
        protected override void InitializeUI()
        {            
            base.InitializeUI();

            this.Text = frmMain.APP_NAME;

            this.saveToolStripMenuItem.Enabled = false;
            this.closeToolStripMenuItem.Enabled = false;

            this.txtCode.KeyDown += new KeyEventHandler(this.txtCode_KeyDown);

            this.parseTreeToolStripMenuItem.Click += new EventHandler(this.OnMenuStripMenuItemClick);
            this.loadGrammarDefinitionToolStripMenuItem.Click += new EventHandler(this.OnMenuStripMenuItemClick);
            this.eBNFEditorToolStripMenuItem.Click += new EventHandler(this.OnMenuStripMenuItemClick);

            this.parseTreeToolStripMenuItem.Enabled = false;
            this.ruleTreeToolStripMenuItem.Enabled = false;
            this.openToolStripMenuItem.Enabled = false;

            this.grammar.ImportComplete -= new GrammarImportCompleteHandler(this.grammar_ImportComplete);
            this.grammar.ImportComplete += new GrammarImportCompleteHandler(this.grammar_ImportComplete);


            int index = this.fileToolStripMenuItem.DropDownItems.IndexOf(this.saveToolStripMenuItem);
            this.fileToolStripMenuItem.DropDownItems.Insert( index, new ToolStripSeparator() );
            this.fileToolStripMenuItem.DropDownItems.Insert( index, this.loadGrammarDefinitionToolStripMenuItem );

            this.viewToolStripMenuItem.DropDownItems.Add( this.parseTreeToolStripMenuItem );

            this.menuStrip1.Items.Add( this.toolsToolStripMenuItem );

            this.RenameMenuItems( this.menuStrip1 );
        }
        #endregion InitializeUI

        #region OnMenuStripMenuItemClick
        private void OnMenuStripMenuItemClick(object sender, EventArgs e)
        {
            ToolStripMenuItem mi = sender as ToolStripMenuItem;

            string name = mi.Name;

            OpenFileDialog ofd = null;
            frmRuleTree ruleTree = null;

            switch ( name )
            {
                case "parseTree":
                    ruleTree = new frmRuleTree();
                    IEnumerable<ParseTreeTokenBase> tokens = this.parseTree.Tokens;
                    ruleTree.ShowMe( this, this.parseTree );
                    break;

                case "loadGrammarDefinition":
                    ofd = new OpenFileDialog();
                    ofd.Multiselect = false;
                    ofd.Filter = frmEBNFEditor.FILE_DIALOG_FILTER;
                    ofd.FileName = string.Empty;
                    ofd.InitialDirectory = frmMain.PATH_GRAMMAR_DEF;
                    if ( ofd.ShowDialog(this) == DialogResult.OK && File.Exists(ofd.FileName) )
                    {
                        bool load = false;
                        ImportResult result = this.LoadGrammarDefinition( ofd.FileName );
                        if ( result.Success == false )
                        {
                            DialogResult dialogResult =
                                MessageBox.Show(
                                "The grammar definition contained errors and could not be loaded. Would you like to open the grammar definition for editing?",
                                "Error",
                                MessageBoxButtons.YesNoCancel,
                                MessageBoxIcon.Error);

                            switch ( dialogResult )
                            {
                                case DialogResult.Cancel:
                                    break;

                                case DialogResult.Yes:
                                    if ( this.OpenEBNFEditor(ofd.FileName) )
                                    {
                                        this.LoadGrammarDefinition( ofd.FileName );
                                        load = true;
                                    }
                                    break;
                            }
                        }
                        else
                        {
                            load = true;
                        }

                        if ( load )
                        {
                            string grammarName = this.grammar.Name.Length > 0 ? this.grammar.Name : frmMain.FilenameFromPath(ofd.FileName, true);
                            this.Text = string.Format( "{0} ({1})", frmMain.APP_NAME, grammarName );
                            this.txtCode.Enabled = true;
                            this.ruleTreeToolStripMenuItem.Enabled = true;
                            this.openToolStripMenuItem.Enabled = true;
                        }
                    }

                    break;

                case "eBNFEditor":
                    
                    this.ClearResults();

                    this.OpenEBNFEditor(this.scriptFilename);
                    
                    //  Apparently the FileSystemWatcher is useless on Vista/Win7
                    //  so just assume the file was changed.
                    if ( string.IsNullOrEmpty(this.scriptFilename) == false )
                    {
                        this.LoadGrammarDefinition(this.scriptFilename);
                        
                        if ( this.highlighter != null )
                            this.highlighter.HighlightText();
                    }

                    this.compileStart = Environment.TickCount;

                    break;
            }
            
        }

        #endregion OnMenuStripMenuItemClick

        #region LoadGrammarDefinition
        private ImportResult LoadGrammarDefinition( string filename )
        {
            if ( this.highlighter != null )
            {
                this.highlighter.Dispose();
                this.highlighter = null;
            }

            this.highlighter = new ParseTreeTextHighlighter( this.txtCode, this.grammar, this.parseTree );
            this.highlighter.ParseComplete += new ParseCompleteHandler(highlighter_ParseComplete);
            this.highlighter.Highlighting += new EventHandler(this.highlighter_Highlighting);
            this.highlighter.HighlightComplete += new EventHandler(this.highlighter_HighlightComplete);

            string s = File.ReadAllText( filename, frmMain.TextEncoding );

            int ticks = Environment.TickCount;

            ImportResult result = this.grammar.Import( s, GrammarDefinitionFormat.EBNF );
            
            ticks = Environment.TickCount - ticks;
            this.SetTimeElapsed( ticks );

            this.Colorize();

            if ( File.Exists(filename) )
                this.scriptFilename = filename;

            this.ClearResults();

            return result;
        }

        #endregion LoadGrammarDefinition

        #region Colorize
        private void Colorize()
        {
            if ( this.grammar.Name.StartsWith("SimpleMath") )
                this.ColorizeSimpleMath();
            else
            if ( this.grammar.Name.StartsWith("C#") )
                this.ColorizeCSharp();
        }

        private void ColorizeCSharp()
        {
            Dictionary<TerminalSymbol, Color> tsColorTable = new Dictionary<TerminalSymbol,Color>();
            Dictionary<SymbolGroup, Color> sgColorTable = new Dictionary<SymbolGroup,Color>();
            foreach ( SymbolGroup sg in this.grammar.SymbolGroups )
            {
                switch ( sg.Key )
                {
                    case "nativeTypes":
                    case "keywords":
                        sgColorTable.Add( sg, Color.Blue );
                        break;
                }
            }


            foreach ( TerminalSymbol ts in this.grammar.TerminalSymbols )
            {
                switch ( ts.Name )
                {
                    case "LINE_COMMENT":
                    case "ANYTHING_COMMENT":
                        tsColorTable.Add( ts, Color.DarkGreen );
                        break;

                    case "TRIPLE_SLASH":
                    case "LT_PLUS_ANYTHING_COMMENT":
                        tsColorTable.Add( ts, Color.DarkGray );
                        break;

                    case "LITERAL_STRING":
                        tsColorTable.Add( ts, Color.Firebrick );
                        break;
                }
            }

            this.highlighter.AssignColorTables( tsColorTable, sgColorTable );
        }

        private void ColorizeSimpleMath()
        {
            Dictionary<TerminalSymbol, Color> colorTable = new Dictionary<TerminalSymbol,Color>();
            foreach ( TerminalSymbol ts in this.grammar.TerminalSymbols )
            {
                switch ( ts.Name )
                {
                    case "ADD":
                    case "SUB":
                        colorTable.Add( ts, Color.Blue );
                        break;

                    case "MUL":
                    case "DIV":
                        colorTable.Add( ts, Color.DarkGreen );
                        break;

                    case "INT":
                    case "FLOAT":
                        colorTable.Add( ts, Color.Magenta );
                        break;

                    default:
                        break;
                }
            }

            this.highlighter.AssignColorTables( colorTable, null );
        }
        #endregion Colorize

        #region OpenEBNFEditor
        private bool OpenEBNFEditor( string filename )
        {
            this.ClearResults();
            frmEBNFEditor editor = new frmEBNFEditor();
            editor.ShowMe( this, filename );
            return editor.HasValidGrammar;
        }
        #endregion OpenEBNFEditor

        #region grammar_ImportComplete
        /// <summary>
        /// Handle the ImportComplete event to set properties of the imported
        /// grammar that is beyiond the scope of an EBNF script.
        /// </summary>
        private void grammar_ImportComplete(object sender, GrammarImportCompleteEventArgs e)
        {
            Grammar g = e.Grammar;

            #region C# language definition
            if ( g.Name.Equals("C# language definition") )
            {
                //  Build a special terminal boundary pattern for numbers,
                //  so that they can be followed by a suffix , and assign that
                //  pattern as the terminal boundary for the number TerminalSymbols.
                string s = g.TerminalSymbols["NUMBER_SUFFIX"].Value;
                string numberSuffixPattern = string.Format("{0}|{1}", s, g.DefaultTerminalBoundaryRegexPattern);
                g.TerminalSymbols["NUMBER_DEC"].Boundary.Pattern = numberSuffixPattern;
                g.TerminalSymbols["NUMBER_INT"].Boundary.Pattern = numberSuffixPattern;

                //  We can use grouping to simplify token colorization.
                SymbolGroup keywords = g.SymbolGroups.Add( "keywords" );
                SymbolGroup nativeTypes = g.SymbolGroups.Add( "nativeTypes" );

                TerminalSymbolsCollection terminalSymbols = g.TerminalSymbols;
                foreach ( TerminalSymbol ts in terminalSymbols )
                {
                    string name = ts.Name;
                    if ( name.EndsWith("_KEYWORD") )
                    {
                        switch ( name )
                        {
                            case "BOOL_KEYWORD":
                            case "BYTE_KEYWORD":
                            case "CHAR_KEYWORD":
                            case "DECIMAL_KEYWORD":
                            case "DOUBLE_KEYWORD":
                            case "FLOAT_KEYWORD":
                            case "OBJECT_KEYWORD":
                            case "SBYTE_KEYWORD":
                            case "SHORT_KEYWORD":
                            case "STRING_KEYWORD":
                            case "UINT_KEYWORD":
                            case "ULONG_KEYWORD":
                            case "USHORT_KEYWORD":
                                ts.Description = "type identifier";
                                ts.Group = nativeTypes;
                                break;

                            case "PRIVATE_KEYWORD":
                            case "PROTECTED_KEYWORD":
                            case "INTERNAL_KEYWORD":
                            case "PUBLIC_KEYWORD":
                                ts.Description = "modifier";
                                ts.Group = keywords;
                                break;

                            default:
                                
                                if ( ts.Comparison != TerminalSymbolComparison.RegularExpression )
                                    ts.Description = ts.Value;
                                
                                ts.Group = keywords;
                                break;
                        }

                    }

                    //  By default only single-character operators are automatically
                    //  assumed to be a terminal parsing boundary, so make any operators
                    //  that are more than one character boundaries as well.
                    if ( ts.Value.Length > 1 &&
                         (name.EndsWith("_OP") ||
                          name.EndsWith("_LOG") ||
                          name.EndsWith("_COND") ||
                          name.EndsWith("_ASSIGN")) )
                        ts.IsBoundary = true;                   
                }

                //  Use required terminal symbols to disambiguate NonTerminalSymbols
                //  that begin with the same terminals, such as property and method
                //  declarations. For example, the string 'public bool Foo' could be
                //  either a property or method declaration, but if the next symbol
                //  is an open parenthesis we know it is a method declaration, so the
                //  open parenthesis is made to be a required symbol for MethodDeclaration,
                //  thus making it disthinguishable from  PropertyDeclaration.
                foreach ( NonTerminalSymbol nts in g.NonTerminalSymbols )
                {
                    switch ( nts.Name )
                    {
                        case "PropertyDeclaration":
                        case "BracedCodeBlock":
                            nts.RequiredTerminalSymbol = g.TerminalSymbols["OPEN_BRACE"];
                            break;

                        case "IndexerDeclaration":
                            nts.RequiredTerminalSymbol = g.TerminalSymbols["THIS_KEYWORD"];
                            break;

                        case "ScopedAssignment":
                            nts.RequiredTerminalSymbol = g.TerminalSymbols["EQUALS_ASSIGN"];
                            break;

                        case "MethodDeclaration":
                        case "MethodReference":
                            nts.RequiredTerminalSymbol = g.TerminalSymbols["OPEN_PAREN"];
                            break;

                        case "Assignment":
                            nts.RequiredTerminalSymbol = g.TerminalSymbols["EQUALS_ASSIGN"];
                            break;

                        case "IfStatement":
                            nts.RequiredTerminalSymbol = g.TerminalSymbols["IF_KEYWORD"];
                            break;
                    }
                }

                //  Add the comment terminals to the PrioritySymbols collection
                //  so they can be parsed wherever they appear.
                g.PrioritySymbols.Add( g.TerminalSymbols["LINE_COMMENT"] );
                g.PrioritySymbols.Add( g.NonTerminalSymbols["TripleSlashComment"] );

                //  Match up the appropriate symbols
                g.AssignTerminalSymbolMatch( g.TerminalSymbols["OPEN_BRACE"], g.TerminalSymbols["CLOSE_BRACE"] );
                g.AssignTerminalSymbolMatch( g.TerminalSymbols["OPEN_PAREN"], g.TerminalSymbols["CLOSE_PAREN"] );
                g.AssignTerminalSymbolMatch( g.TerminalSymbols["OPEN_SQUARE"], g.TerminalSymbols["CLOSE_SQUARE"] );

            }
            #endregion C# language definition
        }
        #endregion grammar_ImportComplete
 
        #region txtCode_KeyDown
        private void txtCode_KeyDown(object sender, KeyEventArgs e)
        {
            RichTextBox textBox = sender as RichTextBox;

            #region Brace matching
            if ( e.Control && (e.KeyData & Keys.E) == Keys.E )
            {
                if ( this.parseTree != null && textBox.SelectionStart >= 0 && textBox.SelectionLength == 0 )
                {
                    ParseNode node = this.parseTree.FindNode(textBox.SelectionStart);
                    TerminalSymbol ts = node != null ? node.Symbol as TerminalSymbol : null;
                    if ( ts != null && ts.MatchingSymbol != null )
                    {
                        node = this.parseTree.FindMatchingNode(node);
                        if ( node != null )
                            textBox.SelectionStart = node.CharacterPosition;
                    }
                }

                e.Handled = true;
            }
            #endregion Brace matching

            #region Selection indenting
            if ( e.KeyCode == Keys.Tab )
            {
                frmMain.HandleTabKey( textBox, this.regexTabs, this.regexLine, e );
            }
            #endregion Selection indenting

            #region Newline auto-indent
            if ( e.KeyCode == Keys.Enter )
            {
                frmMain.HandleEnterKey( textBox, this.regexTabs, this.regexBackToNewline, e );
            }
            #endregion Newline auto-indent
        }
        #endregion txtCode_KeyDown

        #region highlighter_ParseComplete
        private void highlighter_ParseComplete(object sender, ParseCompleteEventArgs e)
        {
            this.parseTree = e.ParseTree;
            IEnumerable<IErrorToken> errors = this.GetErrors(this.parseTree);
            this.LogResults( errors );
            this.toolTip.Enable( errors );

            if ( this.parseTree.HasErrors == false )
                this.parseTreeToolStripMenuItem.Enabled = true;
        }
        #endregion highlighter_ParseComplete

        #region highlighter_Highlighting
        void highlighter_Highlighting(object sender, EventArgs e)
        {
            this.suspended = true;
        }
        #endregion highlighter_Highlighting

        #region highlighter_HighlightComplete
        void highlighter_HighlightComplete(object sender, EventArgs e)
        {
            int ticks = Environment.TickCount - this.compileStart;
            this.SetTimeElapsed( ticks );
            this.suspended = false;
        }
        #endregion highlighter_HighlightComplete

        #region FilenameFromPath
        static internal string FilenameFromPath( string path, bool stripExtension )
        {
            if ( string.IsNullOrEmpty(path) )
                return string.Empty;

            string[] array = path.Split( "\\".ToCharArray() );
            string s = array[ array.Length - 1 ];

            int index = stripExtension ? s.IndexOf(".") : -1;
            return index >= 0 ? s.Substring(0, index) : s;
        }
        #endregion FilenameFromPath

        #region HandleEnterKey
        static private void HandleEnterKey( RichTextBox textBox, Regex regexTabs, Regex regexBackToNewline, KeyEventArgs args )
        {
            int selStart = textBox.SelectionStart;
            if ( selStart < 1 )
                return;

            string s = textBox.Text.Substring( 0, selStart );
            Match m = regexBackToNewline.Match( s );
            if ( m.Success )
            {
                s = textBox.Text.Substring(m.Index, selStart - m.Index);
                m = regexTabs.Match( s );
                if ( m.Success )
                {
                    s = string.Format( "\n{0}", m.Value );
                    textBox.Text = textBox.Text.Insert( selStart, s );
                    textBox.SelectionStart = selStart + m.Length + 1;
                    args.Handled = true;
                    args.SuppressKeyPress = true;
                }
            }
        }

        #endregion HandleEnterKey

        #region HandleTabKey
        static private void HandleTabKey( RichTextBox textBox, Regex regexTabs, Regex regexLine, KeyEventArgs args )
        {
            bool handled = false;
            bool shift = args.Shift;
            string selection = textBox.SelectedText;
            int selStart = textBox.SelectionStart;

            if ( selection.Length > 0 )
            {
                bool hadTabs = false;

                //  Add or remove one tab character depending on whether
                //  the shift key was pressed.
                MatchEvaluator tabEvaluator =
                delegate( Match m )
                {
                    hadTabs = true;

                    if ( shift )
                        return m.Value.Substring(0, m.Length - 1);
                    else
                        return string.Format( "{0}\t", m.Value );
                };

                selection = regexTabs.Replace( selection, tabEvaluator );

                if ( hadTabs == false && shift == false )
                {
                    MatchEvaluator lineEvaluator =
                    delegate( Match m )
                    {
                        return string.Format( "\t{0}", m.Value );
                    };

                    selection = regexLine.Replace( selection, lineEvaluator );
                }
                
                textBox.SelectedText = selection;
                textBox.SelectionStart = selStart;
                textBox.SelectionLength = selection.Length;
                handled = true;
            }
            else
            if ( shift )
            {
                if ( selStart > 0 && textBox.Text[selStart - 1] == '\t' )
                {
                    textBox.Text = textBox.Text.Remove(selStart - 1, 1);
                    textBox.SelectionStart = selStart - 1;
                }

                handled = true;
            }

            if ( handled )
            {
                args.Handled = true;
                args.SuppressKeyPress = true;
            }
        }
        #endregion HandleTabKey

        #region GetErrors
        private IEnumerable<IErrorToken> GetErrors( ParseTree parseTree )
        {
            List<IErrorToken> errors = new List<IErrorToken>(parseTree.Errors.Count);
            foreach( ErrorToken error in parseTree.Errors )
            {
                errors.Add( error );
            }

            return errors;
        }

        #endregion GetErrors

        #region GetLineNumber
        protected override int GetLineNumber( string input, out int column )
        {
            return this.grammar.GetLineNumber( input, out column );
        }
        #endregion GetLineNumber

        #region NewLine
        protected override string NewLine
        {
            get { return this.grammar.NewLine; }
        }
        #endregion NewLine

        #region InitializeOpenFileDialog / InitializeSaveFileDialog
        protected override void InitializeOpenFileDialog( OpenFileDialog dialog )
        {
            dialog.Filter = frmMain.FILE_DIALOG_FILTER;
            dialog.FileName = string.Empty;
            dialog.InitialDirectory = CodeEditorBase.PATH_SOURCE;
        }

        protected override void InitializeSaveFileDialog( SaveFileDialog dialog )
        {
            dialog.Filter = frmMain.FILE_DIALOG_FILTER;
            dialog.FileName = string.Empty;
            dialog.InitialDirectory = CodeEditorBase.PATH_SOURCE;
        }
        #endregion InitializeOpenFileDialog / InitializeSaveFileDialog

        #region OnTextBoxTextChanged
        protected override void OnTextBoxTextChanged()
        {
            this.compileStart = Environment.TickCount;
        }
        #endregion OnTextBoxTextChanged
    }
    #endregion frmMain class
}
