﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Infragistics.SyntaxParsing;

namespace Infragistics.SyntaxParsing.Samples
{
    /// <summary>
    /// Displays a rule tree or parse tree.
    /// </summary>
    public partial class frmRuleTree : Form
    {
        private ParseTree parseTree = null;

        public frmRuleTree()
        {
            this.InitializeComponent();
            this.cmdClose.Click += new EventHandler(cmdClose_Click);
            this.treeView1.NodeMouseDoubleClick += new TreeNodeMouseClickEventHandler(treeView1_NodeMouseDoubleClick);
            this.CancelButton = this.cmdClose;
            this.Text = "Rule Tree";
        }

        void treeView1_NodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if ( this.parseTree != null )
            {
                ParseNode parseNode = e.Node.Tag as ParseNode;
                TokenBase token = parseNode != null ? parseNode.Token : null;

                if ( token != null )
                {
                    frmTokenInfo tokenInfoForm = new frmTokenInfo();
                    tokenInfoForm.ShowMe( this, token );
                }
            }
        }

        private void cmdClose_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        public void ShowMe( Form mainForm, Grammar grammar )
        {
            string name = string.IsNullOrEmpty(grammar.Name) ? null : string.Format(" ({0})", grammar.Name );
            this.Text = string.Format( "Rule tree{0}", name );

            this.PopulateSymbolTree( grammar, this.treeView1 );
            this.ShowDialog( mainForm );
        }

        public void ShowMe( Form mainForm, NonTerminalSymbol nonTerminalSymbol )
        {
            this.Text = string.Format( "Rule tree ({0})", nonTerminalSymbol );

            this.PopulateSymbolTree( nonTerminalSymbol, this.treeView1 );
            this.ShowDialog( mainForm );
        }

        public void ShowMe( Form mainForm, ParseTree parseTree )
        {
            NonTerminalSymbol start = parseTree.Nodes.Count > 0 ? parseTree.Nodes[0].Symbol as NonTerminalSymbol : null;
            string name = start != null ? string.Format(" ({0})", start.Name) : null;;
            this.Text = string.Format( "Parse tree{0}", name );

            this.parseTree = parseTree;
            this.PopulateParseTree( this.parseTree, this.treeView1 );
            this.ShowDialog( mainForm );
        }

        private void PopulateSymbolTree( Grammar grammar, TreeView tree )
        {
            tree.Nodes.Clear();

            if ( grammar.StartSymbol == null )
                return;

            Dictionary<NonTerminalSymbol, object> alreadyProcessed = new Dictionary<NonTerminalSymbol, object>();
            TreeNode node = tree.Nodes.Add( null, string.Format("StartSymbol ('{0}')", grammar.StartSymbol.Name) );
            node.Tag = grammar.StartSymbol;
            alreadyProcessed.Add( grammar.StartSymbol, null );
            frmRuleTree.AddNodes( node.Nodes, grammar.StartSymbol.Rules, alreadyProcessed );

            if ( grammar.PrioritySymbols.Count > 0 )
            {
                alreadyProcessed.Clear();

                TreeNode priority = tree.Nodes.Add( null, "PrioritySymbols" );
                priority.ForeColor = Color.DarkGreen;

                foreach ( Symbol symbol in grammar.PrioritySymbols )
                {
                    NonTerminalSymbol nts = symbol as NonTerminalSymbol;

                    node = priority.Nodes.Add( frmRuleTree.FormatSymbol( symbol ) );
                    node.Tag = symbol;
                    frmRuleTree.SetForeColor( symbol, node );

                    if ( nts != null )
                    {
                        alreadyProcessed.Add( nts, null );
                        frmRuleTree.AddNodes( node.Nodes, nts.Rules, alreadyProcessed );
                    }
                }
            }

            tree.ExpandAll();
        }

        private void PopulateSymbolTree( NonTerminalSymbol nonTerminalSymbol, TreeView tree )
        {
            tree.Nodes.Clear();

            if ( nonTerminalSymbol == null )
                return;

            Dictionary<NonTerminalSymbol, object> alreadyProcessed = new Dictionary<NonTerminalSymbol, object>();
            TreeNode node = tree.Nodes.Add( null, nonTerminalSymbol.Name );
            node.Tag = nonTerminalSymbol;
            alreadyProcessed.Add( nonTerminalSymbol, null );
            frmRuleTree.AddNodes( node.Nodes, nonTerminalSymbol.Rules, alreadyProcessed );

            tree.ExpandAll();
        }

        static private void AddNodes(
            TreeNodeCollection nodes,
            SyntaxRulesCollection rules,
            Dictionary<NonTerminalSymbol, object> alreadyProcessed )
        {
            List<RuleTreePair> subRules = new List<RuleTreePair>();

            foreach ( SyntaxRuleBase rule in rules )
            {
                SyntaxRule sr = rule as SyntaxRule;
                TreeNode node = null;

                node = nodes.Add( frmRuleTree.FormatRule(rule) );
                frmRuleTree.SetForeColor( rule.RuleType, node );

                //  If this rule is a NonTerminalRule and has already been processed,
                //  skip this iteration to (a) avoid infinite recursion and (b) avoid
                //  bloating the size of the TreeView that will be displayed.
                NonTerminalRule ntr = sr != null ? sr as NonTerminalRule : null;
                if ( ntr != null )
                {
                    if ( alreadyProcessed.ContainsKey(ntr.NonTerminalSymbol) )
                        continue;
                    else
                        alreadyProcessed.Add( ntr.NonTerminalSymbol, null );
                }

                if ( sr != null )
                {
                    RuleTreePair pair = new RuleTreePair( node.Nodes, sr.Rules );
                    subRules.Add( pair );
                }
            }

            foreach ( RuleTreePair pair in subRules )
            {
                frmRuleTree.AddNodes( pair.Nodes, pair.Rules, alreadyProcessed );
            }
        }

        static private void SetForeColor( Symbol symbol, TreeNode node )
        {
            SyntaxRuleType ruleType = symbol is TerminalSymbol ? SyntaxRuleType.Terminal : SyntaxRuleType.NonTerminal;
            frmRuleTree.SetForeColor( ruleType, node );
        }

        static private void SetForeColor( SyntaxRuleType ruleType, TreeNode node )
        {
            Color ruleColor = frmRuleTree.GetRuleColor( ruleType );
            if ( ruleColor.IsEmpty == false )
            {
                node.ForeColor = ruleColor;
            }
        }

        static internal Color GetRuleColor( SyntaxRuleType ruleType )
        {
            switch ( ruleType )
            {
                case SyntaxRuleType.Terminal:
                    return Color.Firebrick;

                case SyntaxRuleType.NonTerminal:
                    return Color.Blue;

                default:
                    return Color.FromArgb( 85, 145, 189 );
            }
        }

        static private string FormatRule( SyntaxRuleBase rule )
        {
            TerminalRule terminalRule = rule as TerminalRule;
            NonTerminalRule nonTerminalRule = terminalRule == null ? rule as NonTerminalRule : null;

            if ( terminalRule != null )
                return frmRuleTree.FormatSymbol( terminalRule.TerminalSymbol );
            else
            if ( nonTerminalRule != null )
                return frmRuleTree.FormatSymbol( nonTerminalRule.NonTerminalSymbol );
            else
                return rule.ToString();
        }

        static private string FormatSymbol( Symbol symbol )
        {
            if ( symbol == null )
                return string.Empty;

            TerminalSymbol ts = symbol as TerminalSymbol;
            NonTerminalSymbol nts = symbol as NonTerminalSymbol;

            if ( ts != null )
            {
                string name = ts.Name.Equals(ts.Value) ? "TerminalSymbol" : ts.Name;
                return string.Format( "{0} ('{1}')", name, ts.Value );
            }
            else
                return nts.Name;
        }

        private void PopulateParseTree( ParseTree parseTree, TreeView treeView )
        {
            if ( parseTree == null )
                return;

            treeView.Nodes.Clear();
            this.PopulateParseTree( parseTree.Nodes, treeView.Nodes );
            treeView.ExpandAll();
        }

        private void PopulateParseTree( ParseNodesCollection parseNodes, TreeNodeCollection treeNodes )
        {
            foreach ( ParseNode parseNode in parseNodes )
            {
                Token token = parseNode.Token as Token;
                ErrorToken errorToken = parseNode.Token as ErrorToken;
                NonTerminalRule ntr = parseNode.Rule as NonTerminalRule;

                string text =
                    errorToken != null ?
                    string.Format( "ERROR: InvalidText = '{0}'", errorToken.Text ) :
                    token != null ?
                    string.Format( "'{0}'", token.Text ) :
                    ntr.NonTerminalSymbol.Name;
                    
                TreeNode treeNode = new TreeNode( text );
                treeNode.Tag = parseNode;
                treeNodes.Add( treeNode );

                if ( errorToken != null )
                    treeNode.ForeColor = Color.Red;
                else
                if ( token != null )
                    treeNode.ForeColor = Color.FromArgb( 64, 64, 64 );
                else
                if ( ntr != null )
                {
                    treeNode.ForeColor = Color.FromArgb( 85, 145, 189 );
                    this.PopulateParseTree( parseNode.Nodes, treeNode.Nodes );
                }
            }

        }

        private class RuleTreePair
        {
            public RuleTreePair( TreeNodeCollection nodes, SyntaxRulesCollection rules )
            {
                this.Nodes = nodes;
                this.Rules = rules;
            }

            public SyntaxRulesCollection Rules = null;
            public TreeNodeCollection Nodes = null;
        }
    }
}
