﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;

using Infragistics.SyntaxParsing;

namespace Infragistics.SyntaxParsing.Samples
{
    #region frmStatementCompletion
    /// <summary>
    /// Provides statement completion functionality.
    /// </summary>
    public partial class frmStatementCompletion : Form
    {
        private ItemComparer itemComparer = new ItemComparer();
        private bool hasBeenShown = false;
        private string currentWord = null;
        private int startPos = 0;
        private bool auto = false;
        private List<string> exclusions = null;
        private string additionalChars = string.Empty;
        private bool canceled = false;
        private string lastSelection = null;

        public frmStatementCompletion()
        {
            this.InitializeComponent();

            this.StartPosition = FormStartPosition.Manual;
            this.TopMost = true;
            this.TopLevel = true;

            this.MinimizeBox = false;
            this.MaximizeBox = false;
            this.ControlBox = false;
            this.Text = string.Empty;

            this.lstMembers.KeyDown += new KeyEventHandler(this.lstMembers_KeyDown);
            this.lstMembers.KeyPress += new KeyPressEventHandler(this.lstMembers_KeyPress);
            this.lstMembers.MouseDoubleClick += new MouseEventHandler(this.lstMembers_MouseDoubleClick);

        }

        private void lstMembers_KeyPress(object sender, KeyPressEventArgs e)
        {
            char c = e.KeyChar;
            bool isInputChar = true;

            if ( char.IsWhiteSpace(c) )
                isInputChar = false;
            else
            if ( this.exclusions != null )
            {
                foreach ( string s in this.exclusions )
                {
                    if ( string.IsNullOrEmpty(s) )
                        continue;

                    if ( s[0] == c )
                    {
                        isInputChar = false;
                        break;
                    }
                }
            }

            if ( isInputChar == false )
            {
                this.additionalChars = c.ToString();
                this.SelectAndClose( this.SelectedText );
            }
        }

        private void lstMembers_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            string selected = this.SelectedText;
            this.SelectAndClose( selected );
        }

        private void lstMembers_KeyDown(object sender, KeyEventArgs e)
        {
            bool handled = false;
            string selected = null;

            if ( e.Control || e.Shift || e.Alt )
                handled = true;
            else
            {
                switch ( e.KeyData )
                {
                    case Keys.Enter:
                    case Keys.Tab:
                        selected = this.SelectedText;
                        handled = true;
                        break;

                    case Keys.Escape:
                    case Keys.Back:
                    case Keys.Delete:
                        this.canceled = true;
                        selected = null;
                        handled = true;
                        break;

                    case Keys.Right:
                    case Keys.Left:
                        selected = null;
                        handled = true;
                        break;
                }
            }

            e.Handled = handled;
            e.SuppressKeyPress = handled;

            if ( handled )
                this.SelectAndClose( selected );
        }

        private string SelectedText
        {
            get
            {
                if ( this.canceled )
                    return null;

                ImageListBox.Item selectedItem = 
                    this.lstMembers.SelectedItems.Count > 0 ?
                    this.lstMembers.SelectedItems[0] as ImageListBox.Item :
                    null;

                this.lastSelection = selectedItem != null ? selectedItem.Text : null;
                
                return selectedItem != null ? selectedItem.Text : null;
            }
        }

        private void SelectAndClose( string selected )
        {
            if ( this.canceled == false && this.Closed != null )
            {
                ClosedEventArgs args =
                    new ClosedEventArgs(
                        this.currentWord,
                        selected,
                        this.startPos,
                        this.additionalChars );

                this.Closed.Invoke( this, args );
            }

            this.Close();
        }

        public void ShowMe(
            Form owner,
            List<ImageListBox.Item> items,
            List<string> exclusions,
            string currentWord,
            int startPos,
            bool auto,
            Point location )
        {
            this.exclusions = exclusions;
            this.auto = auto;
            this.currentWord = currentWord;
            this.startPos = startPos;
            this.additionalChars = string.Empty;
            this.canceled = false;

            this.lstMembers.Items.Clear();

            items.Sort( this.itemComparer );

            int initialSelection = -1;

            string startsWith =
                string.IsNullOrEmpty(currentWord) == false ?
                currentWord :
                null;

            foreach ( ImageListBox.Item item in items )
            {
                string text = item.Text;
                if ( string.IsNullOrEmpty(text) )
                    continue;

                if ( string.IsNullOrEmpty(startsWith) == false &&
                     text.StartsWith(startsWith, StringComparison.InvariantCultureIgnoreCase) == false )
                    continue;
                else
                if ( this.lastSelection != null &&
                     this.lastSelection.Equals(text, StringComparison.InvariantCultureIgnoreCase) )
                    initialSelection = this.lstMembers.Items.Count;

                this.lstMembers.Items.Add( item );
            }

            //  If there was only one match, don't show the list, just
            //  fire the event as if that item was selected.
            if ( this.lstMembers.Items.Count == 1 )
            {
                if ( this.Closed != null )
                {
                    ImageListBox.Item item = this.lstMembers.Items[0] as ImageListBox.Item;

                    ClosedEventArgs args =
                        new ClosedEventArgs(
                            this.currentWord,
                            item.Text,
                            this.startPos,
                            this.additionalChars );

                    this.Closed.Invoke( this, args );
                }

                return;
            }

            //  Set the initial selection
            this.lstMembers.SelectedIndex = initialSelection;
            this.lstMembers.TopIndex = initialSelection;

            this.Location = location;

            if ( this.hasBeenShown == false )
            {
                this.hasBeenShown = true;
                this.SetBoundsCore( location.X, location.Y, this.Width, this.Height, BoundsSpecified.Location );
                this.Show( owner );
            }
            else
                this.Visible = true;
        }

        public new event ClosedHandler Closed;

        protected override void OnClosing(CancelEventArgs e)
        {
            base.OnClosing(e);

            if ( this.Disposing == false && this.IsDisposed == false )
            {
                this.Visible = false;
                e.Cancel = true;
            }
        }

        protected override void OnDeactivate(EventArgs e)
        {
            base.OnDeactivate(e);
            this.Close();
        }

    }
    #endregion frmStatementCompletion

    #region ClosedHandler / ClosedEventArgs
    public delegate void ClosedHandler( object sender, ClosedEventArgs e );
    public class ClosedEventArgs
    {
        private string currentText = null;
        private string newText = null;
        private string additionalCharsAfter = null;
        private int startPos = -1;

        internal ClosedEventArgs( string currentText, string newText, int startPos, string additionalCharsAfter )
        {
            this.currentText = currentText;
            this.newText = newText;
            this.startPos = startPos;
            this.additionalCharsAfter = additionalCharsAfter;
        }

        public string CurrentText { get { return string.IsNullOrEmpty(this.currentText) == false ? this.currentText : string.Empty; } }
        public int StartPos { get { return this.startPos; } }
        
        public string Text
        {
            get
            {
                string text = this.newText != null ? this.newText : string.Empty;
                if ( string.IsNullOrEmpty(this.additionalCharsAfter) )
                    return text;

                StringBuilder sb = new StringBuilder();
                
                sb.Append( text );

                if ( string.IsNullOrEmpty(this.additionalCharsAfter) == false )
                    sb.Append( this.additionalCharsAfter );

                return sb.ToString();
            }
        }
    }
    #endregion ClosedHandler / ClosedEventArgs

    #region ItemComparer class
    internal class ItemComparer : IComparer<ImageListBox.Item>
    {
        int IComparer<ImageListBox.Item>.Compare(ImageListBox.Item x, ImageListBox.Item y)
        {
            if ( x == null || y == null )
                return 0;

            return x.Text.CompareTo(y.Text);
        }
    }
    #endregion ItemComparer class

    #region ImageListBox class
    public class ImageListBox : ListBox
    {
        private Timer keyPressTimer = null;
        private StringBuilder keyBuffer = null;

        public ImageListBox()
        {
            this.DrawMode = DrawMode.OwnerDrawFixed;
            this.IntegralHeight = false;
            this.ItemHeight = this.FontHeight + 2;
            this.keyPressTimer = new Timer();
            this.keyPressTimer.Interval = 500;
            this.keyPressTimer.Tick += new EventHandler(this.keyPressTimer_Tick);
            this.keyPressTimer.Enabled = false;
        }

        private void keyPressTimer_Tick(object sender, EventArgs e)
        {
            Timer timer = sender as Timer;
            timer.Stop();
            
            if ( this.keyBuffer != null )
                this.keyBuffer.Remove( 0, this.keyBuffer.Length );
        }

        private void KeySearch( string value )
        {
            Debug.WriteLine( value );

            ObjectCollection items = this.Items;
            int selectedIndex = Math.Max(this.SelectedIndex, 0);
            selectedIndex = Math.Min(selectedIndex, this.Items.Count - 1);

            for ( int loop = 1; loop <= 2; loop += 1 )
            {
                int start = loop == 1 ? selectedIndex : 0;
                int end = loop == 1 ? items.Count - 1 : selectedIndex - 1;
                for ( int i = start; i <= end; i ++ )
                {
                    Item item = items[i] as Item;

                    if ( item.Text.StartsWith(value, StringComparison.InvariantCultureIgnoreCase) )
                    {
                        this.SelectedIndex = i;
                        break;
                    }
                }
            }
        }

        protected override void OnKeyPress(KeyPressEventArgs e)
        {
            base.OnKeyPress(e);

            if ( e.Handled == false && char.IsLetterOrDigit(e.KeyChar) )
            {
                e.Handled = true;

                if ( this.keyBuffer == null )
                    this.keyBuffer = new StringBuilder();

                this.keyBuffer.Append(e.KeyChar);

                this.KeySearch( this.keyBuffer.ToString() );

                this.keyPressTimer.Stop();
                this.keyPressTimer.Start();
            }
        }

        protected override bool IsInputKey(Keys keyData)
        {
            if ( keyData == Keys.Tab )
                return true;

            return base.IsInputKey(keyData);
        }

        protected override void OnDrawItem(DrawItemEventArgs e)
        {
            if ( e.Index < 0 )
            {
                base.OnDrawItem( e );
                return;
            }

            Item item = this.Items[e.Index] as Item;
            Rectangle bounds = e.Bounds;

            Rectangle imageRect = bounds;
            imageRect.Y += 2;
            imageRect.Height -= 4;
            imageRect.X += 1;
            imageRect.Width = bounds.Height - 2;
            using ( SolidBrush brush = new SolidBrush(item.Color) )
            {
                e.Graphics.FillRectangle( brush, imageRect );
            }

            bounds.X += imageRect.Width;
            bounds.Width -= imageRect.Width;

            bool selected = (e.State & DrawItemState.Selected) == DrawItemState.Selected;
            Color backColor = selected ? SystemColors.Highlight : SystemColors.Window;
            Color foreColor = selected ? SystemColors.HighlightText : SystemColors.WindowText;

            Rectangle textBounds = bounds;
            textBounds.X += 2;
            textBounds.Width -= 2;

            using ( SolidBrush brush = new SolidBrush(backColor) )
            {
                e.Graphics.FillRectangle( brush, textBounds );
            }

            using ( SolidBrush brush = new SolidBrush(foreColor) )
            {
                Point textLocation = bounds.Location;
                textLocation.Offset(2, 1);
                e.Graphics.DrawString( item.Text, this.Font, brush, textLocation );
            }
        }

        protected override void Dispose(bool disposing)
        {
            if ( disposing )
            {
            }

            base.Dispose(disposing);
        }

        public class Item
        {
            private string text = null;
            private Color color = Color.Empty;

            public Item ( string text, Color color )
            {
                this.text = text;
                this.color = color;
            }

            public Color Color { get { return this.color; } }
            public string Text { get { return this.text; } }

            public override string ToString()
            {
                return this.Text;
            }
        }
    }
    #endregion ImageListBox class
}
