using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Infragistics.SyntaxParsing;
using Infragistics.SyntaxParsing.Importing;
using System.Drawing.Drawing2D;

namespace SQL_Parsing
{
    public partial class frmMain : Form
    {
        #region Form Constructor
        public frmMain()
        {
            InitializeComponent();
        }
        #endregion

        #region Events
        private void frmMain_Load(object sender, EventArgs e)
        {
            // Handle the two custom events for cursor position and parse complete
            txtEditor.CusorPositionChanged += new RichSyntaxEditor.CursorPositionChangedEventHandler(txtEditor_CusorPositionChanged);
            txtEditor.ParseComplete += new RichSyntaxEditor.ParseCompleteEventHandler(txtEditor_ParseComplete);

            // Add terminal symbol token names and colors for future parsing and colorization
            txtEditor.ColorizedTerminalSymbolTokens.Add("select", Color.Blue);
            txtEditor.ColorizedTerminalSymbolTokens.Add("from", Color.Blue);
            txtEditor.ColorizedTerminalSymbolTokens.Add("where", Color.Blue);
            txtEditor.ColorizedTerminalSymbolTokens.Add("limit", Color.Blue);
            txtEditor.ColorizedTerminalSymbolTokens.Add("or", Color.Blue);
            txtEditor.ColorizedTerminalSymbolTokens.Add("and", Color.Blue);

            // Display the EBNF grammar in the definition text box
            txtEBNFDefinition.Text = Properties.Resources.SQL_EBNF;

            // Import grammar definition
            txtEditor.ImportGrammar(Properties.Resources.SQL_EBNF);

            // Force initial parsing
            txtEditor.ParseEditor();
        }

        void txtEditor_ParseComplete(string output)
        {
            lblRealTimeParsing.Text = output;
        }

        void txtEditor_CusorPositionChanged(string output)
        {
            lblCursorPosition.Text = output;
        }
        #endregion
    }

    #region RichSyntaxEditor (generic inherited control for syntax coloring)
    public class RichSyntaxEditor : RichTextBox
    {
        #region Variables
        private Infragistics.SyntaxParsing.Grammar gmar = new Infragistics.SyntaxParsing.Grammar();
        private const int WM_PAINT = 0x00f;

        private Dictionary<string, Color> _ColorizedTerminalSymbolTokens = new Dictionary<String, Color>();
        /// <summary>
        /// Dictionary of terminal symbol tokens and their font color.
        /// </summary>
        public Dictionary<string, Color> ColorizedTerminalSymbolTokens
        {
            get { return _ColorizedTerminalSymbolTokens; }
            set { _ColorizedTerminalSymbolTokens = value; }
        }

        private Boolean _isPainted = true;
        /// <summary>
        /// Determines whether or not to paint the editor. During the colorization process, you want this set to false.
        /// </summary>
        public Boolean isPainted
        {
            get { return _isPainted; }
            set 
            { 
                _isPainted = value;
                if (value)
                    this.Invalidate();
            }
        }

        /// <summary>
        /// Event used to output the parsing results.
        /// </summary>
        /// <param name="output"></param>
        public delegate void ParseCompleteEventHandler(String output);
        public event ParseCompleteEventHandler ParseComplete;

        /// <summary>
        /// Event used to output the current cursor position.
        /// </summary>
        /// <param name="output"></param>
        public delegate void CursorPositionChangedEventHandler(String output);
        public event CursorPositionChangedEventHandler CusorPositionChanged;
        #endregion

        #region Methods
        public void ImportGrammar(String GrammarDefinition)
        {
            // Import the grammar definition from the application resources
            ImportResult ir = gmar.Import(GrammarDefinition, Infragistics.SyntaxParsing.GrammarDefinitionFormat.EBNF);
            // Any errors? If so, display an error
            if (ir.Errors.Count > 0)
                MessageBox.Show(Properties.Resources.ErrorImportMessage, Properties.Resources.ErrorImportTitle,
                    MessageBoxButtons.OK, MessageBoxIcon.Warning);
        }

        /// <summary>
        /// Forces the parsing of the editor.
        /// </summary>
        public void ParseEditor()
        {
            this.OnTextChanged(new EventArgs());
        }

        private void ProcessTokens(IEnumerable<ParseTreeTokenBase> tokens)
        {
            foreach (ParseTreeTokenBase pttb in tokens)
            {
                if (pttb is ErrorToken)
                {
                    // If token is an error token, color the text red
                    this.Select(pttb.CharacterPosition, pttb.Length);
                    this.SelectionColor = Color.Red;
                    continue;
                }
                else if (pttb.TerminalSymbol == null ||
                    ColorizedTerminalSymbolTokens.ContainsKey(pttb.TerminalSymbol.Value.ToLower()) == false)
                {
                    // If token isn't a terminal symbol and not an error, color it black
                    this.Select(pttb.CharacterPosition, pttb.Length);
                    if (this.SelectionColor != Color.Black)
                        this.SelectionColor = Color.Black;
                    continue;
                }
                else if (ColorizedTerminalSymbolTokens.ContainsKey(pttb.TerminalSymbol.Value.ToLower()) == false)
                {
                    // If token is a defined terminal symbol, but isn't in our dictionary, color it black
                    this.Select(pttb.CharacterPosition, pttb.Length);
                    this.SelectionColor = Color.Black;
                    continue;
                }

                // Token is a terminal symbol and is found in dictionary,
                //   color it based on defined color
                this.Select(pttb.CharacterPosition, pttb.Length);
                this.SelectionColor = (Color)ColorizedTerminalSymbolTokens[pttb.TerminalSymbol.Value.ToLower()];
            }
        }
        #endregion

        #region Processed Events
        /// <summary>
        /// Override the paint method of the control to disallow painting while processing tokens.
        /// </summary>
        /// <param name="m"></param>
        protected override void WndProc(ref System.Windows.Forms.Message m)
        {
            if (m.Msg == WM_PAINT)
            {
                if (isPainted)
                    base.WndProc(ref m);
                else
                    m.Result = IntPtr.Zero;
            }
            else
                base.WndProc(ref m);
        }

        protected override void OnTextChanged(EventArgs e)
        {
            // Copy the current position of the cursor
            int insertionPointPosition = this.SelectionStart;

            // Output the cursor position
            if (this.CusorPositionChanged != null)
                this.CusorPositionChanged(String.Format(Properties.Resources.CursorPosition,
                this.GetLineFromCharIndex(insertionPointPosition) + 1,
                (insertionPointPosition - this.GetFirstCharIndexFromLine(this.GetLineFromCharIndex(insertionPointPosition))) + 1));

            // If the editor has no text, then return
            if (this.Text.Length == 0)
                return;

            this.isPainted = false;

            // Parse the text editor
            ParseTree pt = gmar.Parse(this.Text);

            // Does the Nodes property contain nodes and no have errors?
            if (pt.Nodes != null && pt.Errors.Count == 0)
            {
                DateTime startTime = DateTime.Now;

                // Process the tokens for colorization
                ProcessTokens(pt.Tokens);

                DateTime endTime = DateTime.Now;
                TimeSpan duration = endTime - startTime;

                // Output the parsing result with how long the parsing took
                if (this.ParseComplete != null)
                    this.ParseComplete(String.Format(Properties.Resources.ParsingSuccess, duration));

                // Force update cursor position for info
                this.OnSelectionChanged(new EventArgs());
            }
            else
            {
                // There were errors, but process the tokens
                ProcessTokens(pt.Tokens);
                // Output the parsing result with the error information
                if (this.ParseComplete != null)
                    this.ParseComplete(String.Format(Properties.Resources.ParsingError, pt.Errors[0]));
            }

            // Restore the cursor to the correct location
            this.Select(insertionPointPosition, 0);

            // Set the text color to black
            this.SelectionColor = Color.Black;

            this.isPainted = true;
        }

        protected override void OnSelectionChanged(EventArgs e)
        {
            // If the editor is currently parsing, return
            if (!isPainted)
                return;

            // Copy the current position of the cursor
            int insertionPointPosition = this.SelectionStart;

            // Output the cursor position
            if (this.CusorPositionChanged != null)
                this.CusorPositionChanged(String.Format(Properties.Resources.CursorPosition,
                this.GetLineFromCharIndex(insertionPointPosition) + 1,
                (insertionPointPosition - this.GetFirstCharIndexFromLine(this.GetLineFromCharIndex(insertionPointPosition))) + 1));
        }
        #endregion
    }
    #endregion
}